/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/


#include <windows.h>
#include "CRegistry.h"


// Default constructor
CRegistry::CRegistry()
{
	KeyIsOpen = false ;
}


// Destructor
CRegistry::~CRegistry()
{
	if ( KeyIsOpen == true ) RegCloseKey(hK) ;
}


bool CRegistry::OpenKey(HKEY hKey, const char *subkey, bool create)
{
	unsigned long disp ;

	if ( subkey == NULL ) return false ;
	KeyIsOpen = false ;
	if ( ERROR_SUCCESS == RegOpenKeyEx(hKey, subkey, 0, KEY_ALL_ACCESS, &hK) ) KeyIsOpen = true ;
	else
	{
		if ( create == true )
		{
			if ( ERROR_SUCCESS == RegCreateKeyEx(hKey, subkey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hK, &disp) )
				KeyIsOpen = true ;
		}
	}
	return KeyIsOpen ;
}


bool CRegistry::GetValue(const char *name, long *pVal)
{
	unsigned long val, type, dsize = sizeof(unsigned long);

	if ( ERROR_SUCCESS == RegQueryValueEx(hK, name, NULL, &type, (unsigned char *)&val, &dsize ) )
	{
		if ( REG_DWORD == type )
		{
			*pVal = (long)val ;
			return true ;
		}
		else
			return false ;
	}
	return false ;
}


bool CRegistry::GetValue(const char *name, char *buff, unsigned long bufflen)
{
	unsigned long type, dsize = bufflen ;

	if ( ERROR_SUCCESS == RegQueryValueEx(hK, name, NULL, &type, (unsigned char *)buff, &dsize ) )
		return ( REG_SZ == type ) ? true : false ;
	return false ;
}


bool CRegistry::SetValue(const char *name, long val)
{
	return (ERROR_SUCCESS == RegSetValueEx(hK, name, 0, REG_DWORD, (const unsigned char *)&val, 4)) ? true : false ;
}


bool CRegistry::SetValue(const char *name, const char *val, unsigned long bufflen)
{
	return (ERROR_SUCCESS == RegSetValueEx(hK, name, 0, REG_SZ, (const unsigned char *)val, bufflen)) ? true : false ;
}

