/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/

#include "cconcat.h"

// Default constructor
CConcat::CConcat(void)
{
	StringLength = 0 ;
	Top = 0 ;
	Last = 0 ;
}

// Destructor
CConcat::~CConcat(void){KillString();}


void CConcat::KillString(void)
{

	STRNODE *cnode = Top ;

	if ( StringLength > 0 )
	{
		while ( cnode )
		{
			Last = cnode ;
			cnode = cnode->Next ;
			delete Last->Frag ;
			delete Last ;
		}
		StringLength = 0 ;
		Top = 0 ;
		Last = 0 ;
	}
}

// Return length of a NULL terminated string
unsigned long CConcat::cc_strlen(const char *Str)
{
  const char *p ;

  p = Str ;
  while ( *p ) p++ ;
  return (p - Str) ;
}


// Copy a NULL terminated string
char *CConcat::cc_strdup(const char *StrToDup)
{
  unsigned long l ;
  const char *p ;
  char *q, *r ;

  l = cc_strlen(StrToDup) ;
  r = new char[l+1] ;
  p = StrToDup ;
  q = r ;
  while ( *p ) *q++ = *p++ ; 
  *q = 0 ;
  return r ;
}


// Add a new node to the linked list for the next string fragment
void CConcat::Append(const char *Fragment)
{
	STRNODE *tmpnode = new STRNODE ;

	tmpnode->FragLen = cc_strlen(Fragment) ;
	StringLength += tmpnode->FragLen ;
	tmpnode->Frag = cc_strdup(Fragment) ;
	tmpnode->Next = 0 ;
	if ( Top == 0 ) 
	{
		Top = tmpnode ;
		Last = Top ;
	}
	else
	{
		Last->Next = tmpnode ;
		Last = tmpnode ;
	}
}


// Output the resulting string
char *CConcat::ToString(void)
{
	STRNODE *cnode = Top ;
	char *result ;
	char *p, *q ;
	unsigned long i ;

	if ( StringLength == 0 ) return 0 ;
	result = new char[StringLength+1] ;
	p = result ;
	do
	{
		for ( i = 0, q = cnode->Frag ; i < cnode->FragLen ; i++, *p = 0 )
			*p++ = *q++ ;
		cnode = cnode->Next ;
	}	while ( cnode ) ;

	return result ;
}

void CConcat::Reset(void){KillString();}

