/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/

/*
 * Based on the Scintilla sample app
 */

#include <stdio.h>
#include "cscre.h"
//#include "scrkwds.h"
#include "scredit.h"
#include "xmlparse.h"

#include <commctrl.h>		// Needed for PBM_SETBKCOLOR and toolbar/status bar stuff

extern HWND hwMDIFrame, hwToolBar, hwStatusBar ;
extern long bgColour, fgColour, kwColour, paColour, acColour, faColour, syColour ;
extern long coColour, nuColour, qsColour, ukColour ;
extern long AllowAutoC ;
extern int ac_on ;
extern char *tmpacw ;
extern char *AoCKeyWords, *ActionKeyWords, *ParamKeyWords, *FactKeyWords, *SysObjectWords, *acWords, *UserKeyWords ;
extern XML ScriptXML ;
extern bool TipsSwitchedOn ;
extern long cmdFontSize, comFontSize ;
extern char cmdFont[FONT_NAME_SIZE], comFont[FONT_NAME_SIZE] ;
//extern SECOLOUR Colours[] ;
extern EDWIN *Active ;

// Constructor
ScrEApp::ScrEApp()
{
	hInstance = 0;
	currentDialog = 0;
	wMain = 0;
	wEditor = 0;
	isDirty = false;
	fullPath[0] = '\0';
	acBuff[0] = 0 ;
	acbp = acBuff ;
	tips_on = false ;
}


// Destructor
ScrEApp::~ScrEApp(void) {}


void ScrEApp::GetRange(int start, int end, char *text)
{
	TEXTRANGE tr;


	tr.chrg.cpMin = start;
	tr.chrg.cpMax = end;
	tr.lpstrText = text;
	SendMessage(wEditor, EM_GETTEXTRANGE, 0, reinterpret_cast<LPARAM>(&tr));
}


void ScrEApp::SetTitle()
{
	char title[MAX_PATH + 100];


	strcpy(title, appName);
	strcat(title, " - ");
	strcat(title, fullPath);
	::SetWindowText(wMain, title);
}


void ScrEApp::New()
{
	SendEditor(SCI_CLEARALL);
	SendEditor(EM_EMPTYUNDOBUFFER);
	fullPath[0] = '\0';
	SetTitle();
	isDirty = false;
	SendEditor(SCI_SETSAVEPOINT);
}


void ScrEApp::OpenFile(const char *fileName)
{
	New();
	SendEditor(SCI_CANCEL);
	SendEditor(SCI_SETUNDOCOLLECTION, 0);
	strcpy(fullPath, fileName);
	FILE *fp = fopen(fullPath, "rb");
	if (fp)
	{
		SetTitle();
		char data[blockSize];
		int lenFile = fread(data, 1, sizeof(data), fp);
		while (lenFile > 0)
		{
			SendEditor(SCI_ADDTEXT, lenFile,  reinterpret_cast<LPARAM>(static_cast<char *>(data)));
			lenFile = fread(data, 1, sizeof(data), fp);
		}
		fclose(fp);
	}
	else
	{
		char msg[MAX_PATH + 100];
		strcpy(msg, "Could not open file \"");
		strcat(msg, fullPath);
		strcat(msg, "\".");
		::MessageBox(wMain, msg, appName, MB_OK);
	}
	SendEditor(SCI_SETUNDOCOLLECTION, 1);
	::SetFocus(wEditor);
	SendEditor(EM_EMPTYUNDOBUFFER);
	SendEditor(SCI_SETSAVEPOINT);
	SendEditor(SCI_GOTOPOS, 0);
}


void ScrEApp::Save(){ SaveFile(fullPath); }


void ScrEApp::SaveFile(const char *fileName)
{
	FILE *fp = fopen(fullPath, "wb");


	if (fp)
	{
		char data[blockSize + 1];
		int lengthDoc = SendEditor(SCI_GETLENGTH);
		for (int i = 0; i < lengthDoc; i += blockSize)
		{
			int grabSize = lengthDoc - i;
			if (grabSize > blockSize)	grabSize = blockSize;
			GetRange(i, i + grabSize, data);
			fwrite(data, grabSize, 1, fp);
		}
		fclose(fp);
		SendEditor(SCI_SETSAVEPOINT);
	}
	else
	{
		char msg[MAX_PATH + 100];
		strcpy(msg, "Could not save file \"");
		strcat(msg, fullPath);
		strcat(msg, "\".");
		MessageBox(wMain, msg, appName, MB_OK);
	}
}


int ScrEApp::SaveIfUnsure()
{
	if (isDirty)
	{
		char msg[MAX_PATH + 100];
		strcpy(msg, "Save changes to \"");
		strcat(msg, fullPath);
		strcat(msg, "\"?");
		int decision = MessageBox(wMain, msg, appName, MB_YESNOCANCEL);
		if (decision == IDYES) Save();
		return decision;
	}
	return IDYES;
}




/*
void ScrEApp::Command(int id)
{
  switch (id)
  {
    case IDM_FILE_NEW:
      if (SaveIfUnsure() != IDCANCEL) New();
      break;

    case IDM_FILE_OPEN:
      if (SaveIfUnsure() != IDCANCEL) Open();
      break;

    case IDM_FILE_EXIT:
      if (SaveIfUnsure() != IDCANCEL) ::PostQuitMessage(0);
      break;
	}
}
*/

void ScrEApp::EnableAMenuItem(int id, bool enable)
{
	if (enable)
	{
		::EnableMenuItem(::GetMenu(hwMDIFrame), id, MF_ENABLED | MF_BYCOMMAND);
		SendMessage(hwToolBar, TB_ENABLEBUTTON, id, TRUE) ;
	}
	else
	{
		::EnableMenuItem(::GetMenu(hwMDIFrame), id, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
		SendMessage(hwToolBar, TB_ENABLEBUTTON, id, FALSE) ;
	}
}


void ScrEApp::CheckMenus(void)
{
  EnableAMenuItem(IDM_FILE_SAVE, isDirty);
  EnableAMenuItem(IDM_EDIT_UNDO, SendEditor(EM_CANUNDO) == 0 ? false : true);
  EnableAMenuItem(IDM_EDIT_REDO, SendEditor(SCI_CANREDO) == 0 ? false : true);
  EnableAMenuItem(IDM_EDIT_PASTE, SendEditor(EM_CANPASTE) == 0 ? false : true);
}


void ScrEApp::Notify(SCNotification *notification)
{

	char *p, *q, *r, tmpacbuff[MAX_PATH+1], tipbuff[MAX_PATH+1] ;
	TextToFind tt ;


	switch (notification->nmhdr.code)
	{
		case SCN_PAINTED:
		if ( Active->app != this ) break ;
		{
			char sbinfo[SB_ITEM_LENGTH+1] ;
			long curpos = SendEditor(SCI_GETCURRENTPOS) ;
			long bytes = SendEditor(SCI_GETLENGTH) ;
			long lines = SendEditor(SCI_GETLINECOUNT) ;
			long cline = SendEditor(SCI_LINEFROMPOSITION, curpos) ;
			sprintf(sbinfo, "%ld/%ld", cline+1, lines) ;
			SendMessage(hwStatusBar, SB_SETTEXT, 0, (LPARAM)sbinfo) ;
			sprintf(sbinfo, "%ld", bytes) ;
			SendMessage(hwStatusBar, SB_SETTEXT, 1, (LPARAM)sbinfo) ;
		}
			break ;

		case SCN_SAVEPOINTREACHED:
			isDirty = false;
			CheckMenus();
			break;

		case SCN_SAVEPOINTLEFT:
			isDirty = true;
			CheckMenus();
			break;

		case SCN_DWELLSTART:
			if (TipsSwitchedOn == false) break ;
			SendEditor(SCI_CALLTIPCANCEL) ;
			if (tips_on == true ) break ;
			if ( SendEditor(SCI_CALLTIPACTIVE) == 0 )
			{
				tt.chrg.cpMax = SendEditor(SCI_WORDENDPOSITION, notification->position, 1) ;
				tt.chrg.cpMin = SendEditor(SCI_WORDSTARTPOSITION, notification->position, 1) ;
				tt.lpstrText = tipbuff ;
				SendEditor(SCI_GETTEXTRANGE, 0, (long)&tt) ;
				if ( strstr((const char *)acWords, (const char *)tipbuff) != NULL )
				{
					if (ScriptXML.FindNode("keyword","name",(const char *)tipbuff) == 0)
					{
						tip_name = ScriptXML.GetNodeAttrValue("name") ;
						tip_cat = ScriptXML.GetNodeAttrValue("category") ;
						tip_subcat = ScriptXML.GetNodeAttrValue("subcategory") ;
						ScriptXML.MoveToFirstChild() ;
						tip_usage = ScriptXML.NodeValue() ;
						char sbinfo[SB_ITEM_LENGTH+1] ;
						if ( strlen(tip_usage) <= SB_ITEM_LENGTH )
							sprintf(sbinfo,"%s",tip_usage) ;
						else
							strncpy(sbinfo, tip_usage, SB_ITEM_LENGTH ) ;
						sbinfo[SB_ITEM_LENGTH] = 0 ;
						SendMessage(hwStatusBar, SB_SETTEXT, 3, (LPARAM)sbinfo) ;
						ScriptXML.MoveToNextSibling() ;
						strcpy(CallTip, (const char *)ScriptXML.NodeValue()) ;
						if (strcmp((const char *)CallTip, "NULL") == 0) break ;
						p = CallTip ;
						while (*p) 
						{
							if ( *p == '\\' && *(p+1) == 'n')
							{
								*p = 32 ;
								*(p+1) = 10 ;
							}
							p++ ;
						}
						tip_desc = CallTip ;
						SendEditor(SCI_CALLTIPSHOW,
									notification->position, 
									reinterpret_cast<LPARAM>(CallTip)) ;
						tips_on = true ;
					}
				}
			}
			break;

		case SCN_DWELLEND:
			tips_on = false ;
			SendMessage(hwStatusBar, SB_SETTEXT, 3, (LPARAM)"") ;

//			if ( SendEditor(SCI_CALLTIPACTIVE) == 1 )
//				SendEditor(SCI_CALLTIPCANCEL) ;
			break;

		case SCN_CALLTIPCLICK:
			tips_on = true ;
			SendEditor(SCI_CALLTIPCANCEL) ;
			sprintf(CallTipMsg, "Keyword: %s\nCategory: %s/%s\nUsage: %s\nDescription:\n%s",
					tip_name, tip_cat, tip_subcat, strcmp(tip_usage,"NULL") == 0 ? "not available": tip_usage, tip_desc) ;
			MessageBox(hwMDIFrame, CallTipMsg, "Keyword Information", MB_OK) ;
			tips_on = false ;
			break ;

		case SCN_MODIFIED:
			if ( notification->modificationType & SC_MOD_DELETETEXT )
			{  // Can't trap backspace, so use deleted chars notification
				if ( notification->length == 1)
				{
					if ( acbp > acBuff )
						*--acbp = 0 ;
				}
				else
				{ // More than 1 char deleted, so reset auto-complete
					tt.chrg.cpMax = SendEditor(SCI_WORDENDPOSITION, notification->position, 1) ;
					tt.chrg.cpMin = SendEditor(SCI_WORDSTARTPOSITION, notification->position, 1) ;
					tt.lpstrText = acBuff ;
					SendEditor(SCI_GETTEXTRANGE, 0, (long)&tt) ;

					acbp = acBuff + (tt.chrg.cpMax - tt.chrg.cpMin) ;
					*acbp = 0 ;
				}
			}
			break ;

		case SCN_CHARADDED:
			switch( notification->ch)
			{
				case ';':
					ac_on = 0 ;
					break ;

				case '"':
					ac_on = (ac_on == 0)? 1:0 ;
					break ;

				case 13:
				case 10:
					ac_on = 1 ;
				case 32:
				case '(':
				case ')':
				case 9:
					acbp = acBuff ;
					acBuff[0] = 0 ;
					break ;

				default:
					if ( acbp - acBuff < MAX_PATH-5 && ac_on == 1 )
					{
//						*acbp++ = notification->ch ;
//						*acbp = 0 ;

						tt.chrg.cpMax = SendEditor(SCI_WORDENDPOSITION, SendEditor(SCI_GETCURRENTPOS), 1) ;
						tt.chrg.cpMin = SendEditor(SCI_WORDSTARTPOSITION, SendEditor(SCI_GETCURRENTPOS), 1) ;
						if ( tt.chrg.cpMax - tt.chrg.cpMin > MAX_PATH ) break ;
						tt.lpstrText = acBuff ;
						SendEditor(SCI_GETTEXTRANGE, 0, (long)&tt) ;
						acbp = acBuff + (tt.chrg.cpMax - tt.chrg.cpMin) ;
						*acbp = 0 ;

						if ( acbp - acBuff > 2 && AllowAutoC == BST_CHECKED)
						{
							strcpy(tmpacw, (const char *)acWords) ;
							tmpacbuff[0] = 32 ;
							strcpy(tmpacbuff+1, (const char *)acBuff) ;
							p = strstr(tmpacw, (const char *)tmpacbuff) ;
							if ( p != NULL )
							{
								r = p + 1 ;
								do
								{
									q = strstr((const char *)r, (const char *)tmpacbuff) ;
									if ( q != NULL ) r = q + 1 ;

								} while ( q != NULL ) ;							
								while ( *r && *r != 32 ) r++ ;
								if ( *r == 32 ) *r = 0 ;
								SendEditor(SCI_AUTOCSHOW, acbp - acBuff, (long)(p+1) ) ;
							}
						}
					}
					break ;
			}
			break ;
	}
}



void ScrEApp::SetAStyle(int style, COLORREF fore, COLORREF back, int size, const char *face)
{
  SendEditor(SCI_STYLESETFORE, style, fore);
  SendEditor(SCI_STYLESETBACK, style, back);
  if (size >= 1) SendEditor(SCI_STYLESETSIZE, style, size);
  if (face) SendEditor(SCI_STYLESETFONT, style, reinterpret_cast<LPARAM>(face));
}


void ScrEApp::InitialiseEditor() 
{

	SendEditor(SCI_SETLEXER, SCLEX_AOC);
	SendEditor(SCI_SETSTYLEBITS, 7);

    SendEditor(SCI_AUTOCSETMAXHEIGHT, 7) ;

	SendEditor(SCI_SETKEYWORDS, 0, 
		reinterpret_cast<LPARAM>(AoCKeyWords));
	SendEditor(SCI_SETKEYWORDS, 1, 
		reinterpret_cast<LPARAM>(ParamKeyWords));
	SendEditor(SCI_SETKEYWORDS, 2, 
		reinterpret_cast<LPARAM>(ActionKeyWords));
	SendEditor(SCI_SETKEYWORDS, 3, 
		reinterpret_cast<LPARAM>(FactKeyWords));
	SendEditor(SCI_SETKEYWORDS, 4, 
		reinterpret_cast<LPARAM>(SysObjectWords));
	if ( UserKeyWords != NULL )
		SendEditor(SCI_SETKEYWORDS, 5, 
			reinterpret_cast<LPARAM>(UserKeyWords));

	SendEditor(SCI_SETWORDCHARS, 0, 
		reinterpret_cast<LPARAM>("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-#"));
	SendEditor(SCI_SETMOUSEDWELLTIME, DWELL_TIME) ;

	// Set up the global default style. These attributes are used wherever no explicit choices are made.
	SetAStyle(STYLE_DEFAULT, fgColour, bgColour, cmdFontSize, cmdFont);
	SendEditor(SCI_STYLECLEARALL);	// Copies global style to all others

	SetAStyle(SCE_AOC_DEFAULT, fgColour, bgColour, cmdFontSize, cmdFont);
	
	SetAStyle(SCE_AOC_WORD, kwColour);		// 0 - Keywords 
	SetAStyle(SCE_AOC_WORD2, paColour);		// 1 - Params 
	SetAStyle(SCE_AOC_WORD3, acColour);		// 2 - Actions
	SetAStyle(SCE_AOC_WORD4, faColour);		// 3 - Facts
	SetAStyle(SCE_AOC_WORD5, syColour);		// 4 - System symbols

	SetAStyle(SCE_AOC_WORD6, ukColour);			// 5 - User keywords

	SetAStyle(SCE_AOC_DEFAULT, fgColour);		// Might need reg val for this (def=black)
	SetAStyle(SCE_AOC_COMMENTLINE, coColour);	// Comments
	SetAStyle(SCE_AOC_NUMBER, nuColour);		// Numbers
	SendEditor(SCI_STYLESETBOLD, SCE_AOC_WORD, 1);
	SetAStyle(SCE_AOC_STRING, qsColour);		// Quoted strings
	SetAStyle(SCE_AOC_IDENTIFIER, fgColour);	// Might need reg val for this (def=black)
	
    // set all items to be in Courier on a white background
	for (int bstyle=SCE_AOC_DEFAULT; bstyle<=SCE_AOC_WORD8; bstyle++) 
    {
		SendEditor(SCI_STYLESETFONT, bstyle, reinterpret_cast<LPARAM>(cmdFont));
		SendEditor(SCI_STYLESETBACK, bstyle, bgColour);
		SendEditor(SCI_STYLESETEOLFILLED, bstyle, 1);
	}
//	SendEditor(SCI_STYLESETBACK, SCE_AOC_STRINGEOL, RGB(0x7F,0x7F,0xFF));

    // Display comments in Comic Sans Serif
	SendEditor(SCI_STYLESETFONT, SCE_AOC_COMMENTLINE, reinterpret_cast<LPARAM>(comFont));
	SendEditor(SCI_STYLESETSIZE, SCE_AOC_COMMENTLINE, comFontSize) ;

	if ( fnbuff[0] != 0 )
	{
		if ( fnbuff[0] == '"' )
		{
			char *p = fnbuff+2 ;
			while (*p && *p != '"') p++ ;
			*p = 0 ;
			OpenFile((const char *)(fnbuff+1)) ;
		}
		else OpenFile((const char *)fnbuff) ;
	}
	fnbuff[0] = 0 ;
	SendEditor(SCI_COLOURISE,0,-1) ;
}
