/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/

/*
 * Based (loosely) on the Scintilla sample app
 */


/*
 * Standard operating system includes
 */

#include <direct.h>			// Needed only for _getcwd()
#include <stdio.h>
#include <fcntl.h>			// Needed only for definition of _O_BINARY
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

/*
 * Application includes
 */
#include "cscre.h"			// Scintilla wrapper class
#include "CRegistry.h"		// Wrapper class for Windows registry functions
#include "scredit.h"		// Main module header
#include "xmlparse.h"		// XML parser class
#include "cconcat.h"		// Big string concatenation class

#include <commctrl.h>		// Needed for PBM_SETBKCOLOR and toolbar/status bar stuff


/*
 * Global variables
 */

CColourPicker *cpicker = NULL ;
XML			ScriptXML ;
HWND		hwMDIFrame, hwMDIClient, hwCurrentDialogue = NULL, hwRefDialogue = NULL ;
HWND		hwToolBar, hwStatusBar, hwToolTip ;
HWND		hwHelp = NULL, hwKW = NULL, hwCat = NULL, hwSubCat = NULL, hwEdit = NULL, hwBut = NULL ;
HINSTANCE	hInst ;
EDWIN		*Top = NULL, *Active = NULL, *Last = NULL ;
char		fnbuff[MAX_PATH] ;
long		bgColour, fgColour, kwColour, paColour, acColour, faColour, syColour ;
long		coColour, nuColour, qsColour, ukColour ;
long		AllowAutoC, ColourConv = 0 ;
int			LineToGoto, LastCat = 0, LastSubCat = 0, UserKWCount = 0 ;
char		fText[MAX_PATH], rText[MAX_PATH] ;
int			ac_on = 1 ;
TextToFind	ttf ;
bool		dlgFocus = false, dlgCancel = false ;
char		*AoCKeyWords, *ActionKeyWords, *ParamKeyWords, *FactKeyWords, *SysObjectWords, *acWords, *UserKeyWords ;
char		*tmpacw ;
bool		TipsSwitchedOn, gotmdi = false ;
long		cmdFontSize, comFontSize ;
char		cmdFont[FONT_NAME_SIZE], comFont[FONT_NAME_SIZE] ;
const char	**RefCat, **RefSubCat ;
char		RefDesc[REF_DESC_SIZE] ;
RECENTFILE	Recent[MAX_RECENT] ;
RECENTFILE	*rfptr[MAX_RECENT] ;
long		rfcount ;
char		work[WORK_BUFF_SIZE] ;
ACWORD		*ACWTop = NULL, *ACWLast = NULL ;
CConcat		acwords, usrwords ;
PBINFO		pbinf[MAX_COLOURS] ;

// v2.1.3+ need these for backwards compatibility
SECOLOUR Colours[] =
{

	{"Black",		RGB(	0,		0,		0)		},		// 00
	{"White",		RGB(	0xff,	0xff,	0xff)	},		// 01
	{"Red",			RGB(	0xff,	0x10,	0x10)	},		// 02
	{"Blue",		RGB(	0x10,	0x10,	0xff)	},		// 03
	{"Green",		RGB(	0xff,	0x10,	0x10)	},		// 04
	{"Off-White",	RGB(	0xff,	0xfb,	0xf0)	},		// 05
	{"Light Grey",	RGB(	0xa0,	0xa0,	0xa0)	},		// 06
	{"Medium Grey",	RGB(	0x80,	0x80,	0x80)	},		// 07
	{"Dark Grey",	RGB(	0x20,	0x20,	0x20)	},		// 08
	{"Dark Green",	RGB(	0x07,	0x80,	0x07)	},		// 09
	{"Dark Blue",	RGB(	0x07,	0x07,	0x80)	},		// 10
	{"Dark Red",	RGB(	0x80,	0x07,	0x07)	},		// 11
	{"Light Blue",	RGB(	0x40,	0x40,	0xf0)	},		// 12
	{"Light Green",	RGB(	0x40,	0xf0,	0x40)	},		// 13
	{"Light Red",	RGB(	0xf0,	0x40,	0x40)	},		// 14
	{"Yellow",		RGB(	0xc0,	0xc0,	0x10)	},		// 15
	{"Cyan",		RGB(	0x10,	0xc0,	0xc0)	},		// 16
	{"Magenta",		RGB(	0xc0,	0x10,	0xc0)	},		// 17
	{"Brown",		RGB(	0x80,	0x60,	0x10)	},		// 18
	{"Aquamarine",	RGB(	0x07,	0x80,	0x80)	},		// 19
	{"Light Magenta",RGB(	0x80,	0x07,	0xc8)	},		// 20
	{"Rice Yellow",	RGB(	0xf6,	0xf7,	0xda)	},		// 21
	{"Dull Gold",	RGB(	0xda,	0xdb,	0xc3)	},		// 22
// Add new colours between here

// And here.
	{NULL,			0}
} ;


void SetRegRecentFiles(void)
{
	CRegistry reg ;
	int i ;
	char buff[20] ;
	unsigned long l ;


	if ( true == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY) )
	{
		for( i = 0 ; i < rfcount ; i++ )
		{
			sprintf(buff, "recent%d", i+1) ;
			l = strlen((const char *)rfptr[i]->FullPath) ;
			reg.SetValue((const char *)buff, (const char *)rfptr[i]->FullPath, l) ;
		}
		reg.SetValue("rfcount", rfcount) ;
	}
}


void GetRegRecentFiles(void)
{
	CRegistry reg ;
	long i, len ;
	char buff[20], *p ;


	if ( true == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY) )
	{
		reg.GetValue("rfcount", &rfcount) ;
		for ( i = 0 ; i < rfcount ; i++ )
		{
			sprintf(buff, "recent%d", i+1) ;
			reg.GetValue((const char *)buff, Recent[i].FullPath, MAX_PATH) ;
			Recent[i].id = IDM_FILESTART + i + 1 ;
			Recent[i].ed = NULL ;
			rfptr[i] = &Recent[i] ;
			if ( (len = strlen(Recent[i].FullPath)) <= MAX_DISPLAY_SIZE - 5)
				sprintf(Recent[i].DisplayName, "%d. %s", i, Recent[i].FullPath) ;
			else
			{
				p = Recent[i].FullPath + len - MAX_DISPLAY_SIZE + 7 ;
				sprintf(Recent[i].DisplayName, "%d. ...%s", i+1, p) ;
			}
			AppendMenu(	GetSubMenu(GetMenu(hwMDIFrame), FILEMENU),
						MF_STRING|MF_UNCHECKED|MF_ENABLED,
						Recent[i].id, Recent[i].DisplayName) ;
		}
		DrawMenuBar(hwMDIFrame) ;
	}
}


/*
 * Set registry settings. These are mainly items from 
 * the settings dialogue box.
 */
void SetRegistrySettings(void)
{
	CRegistry reg ;
	unsigned long l ;


	if ( true == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY) )
	{
		reg.SetValue("bgColour", bgColour) ;
		reg.SetValue("fgColour", fgColour) ;
		reg.SetValue("kwColour", kwColour) ;
		reg.SetValue("paColour", paColour) ;
		reg.SetValue("acColour", acColour) ;
		reg.SetValue("faColour", faColour) ;
		reg.SetValue("syColour", syColour) ;
		reg.SetValue("coColour", coColour) ;
		reg.SetValue("nuColour", nuColour) ;
		reg.SetValue("qsColour", qsColour) ;
		reg.SetValue("ukColour", ukColour) ;
		reg.SetValue("AutoComplete", AllowAutoC) ;
		reg.SetValue("cmdFontSize", cmdFontSize) ;
		reg.SetValue("comFontSize", comFontSize) ;
		l = strlen((const char *)cmdFont) ;
		reg.SetValue("cmdFont", (const char *)cmdFont, l) ;
		l = strlen((const char *)comFont) ;
		reg.SetValue("comFont", (const char *)comFont, l) ;
	}
}


/*
 * Set registry settings. Also sets default values if this is the
 * first time we have been run.
 */
void GetRegistrySettings(void)
{
	CRegistry reg ;


	if ( false == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY) )
	{
		if ( false == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY, true) )
			MessageBox(hwMDIFrame,"Permissions problem writing to registry", "Permission Denied", MB_OK) ;
		else
		{
			// Set up default values
			bgColour = 1 ;
			fgColour = 0 ;
			kwColour = 10 ;
			paColour = 2 ;
			acColour = 9 ;
			faColour = 12 ;
			syColour = 18 ;
			coColour = 7 ;
			nuColour = 19 ;
			qsColour = 20 ;
			ukColour = 0 ;
			cmdFontSize = 10 ;
			comFontSize = 10 ;
			AllowAutoC = BST_CHECKED ;
			sprintf(cmdFont, "Courier New") ;
			sprintf(comFont, "Comic Sans MS") ;
			SetRegistrySettings() ;
			reg.SetValue("rfcount", 0) ;
			ColourConv = 0 ;
		}
	}
	else
	{
		reg.GetValue("bgColour", &bgColour) ;
		reg.GetValue("fgColour", &fgColour) ;
		reg.GetValue("kwColour", &kwColour) ;
		reg.GetValue("paColour", &paColour) ;
		reg.GetValue("acColour", &acColour) ;
		reg.GetValue("faColour", &faColour) ;
		reg.GetValue("syColour", &syColour) ;
		reg.GetValue("coColour", &coColour) ;
		reg.GetValue("nuColour", &nuColour) ;
		reg.GetValue("qsColour", &qsColour) ;
		ukColour = 0 ; // might not be set in registry
		reg.GetValue("ukColour", &ukColour) ;
		reg.GetValue("AutoComplete", &AllowAutoC) ;
		reg.GetValue("cmdFontSize", &cmdFontSize) ;
		reg.GetValue("comFontSize", &comFontSize) ;
		reg.GetValue("cmdFont", cmdFont, FONT_NAME_SIZE) ;
		reg.GetValue("comFont", comFont, FONT_NAME_SIZE) ;
		reg.GetValue("ColourConv", &ColourConv) ;
	}
	if ( ColourConv == 0 )
	{
		bgColour = Colours[bgColour].colour ;
		fgColour = Colours[fgColour].colour ;
		kwColour = Colours[kwColour].colour ;
		paColour = Colours[paColour].colour ;
		acColour = Colours[acColour].colour ;
		faColour = Colours[faColour].colour ;
		syColour = Colours[syColour].colour ;
		coColour = Colours[coColour].colour ;
		nuColour = Colours[nuColour].colour ;
		qsColour = Colours[qsColour].colour ;
		ukColour = Colours[ukColour].colour ;
		ColourConv = 1 ;
		reg.SetValue("ColourConv", ColourConv) ;
		SetRegistrySettings() ;
	}
}


void UpdateFileMenu(const char *fname, EDWIN *ed)
{
	RECENTFILE *rf ;
	char *p ;
	int i, len, pos = -1 ;


	for (i = 0 ; i < rfcount ; i++ )
	{
		if ( strcmp((const char *)rfptr[i]->FullPath, fname) == 0 ) pos = i ;
		RemoveMenu(GetSubMenu(GetMenu(hwMDIFrame), FILEMENU), rfptr[i]->id, MF_BYCOMMAND) ;
	}

	if ( pos == -1 )
	{
		if ( rfcount < MAX_RECENT )
		{
			rf = &Recent[rfcount] ;
			rfcount++ ;
		}
		else
			rf = rfptr[MAX_RECENT-1] ;
	}
	else
		rf = rfptr[pos] ;

	rf->ed = ed ;
	strcpy(rf->FullPath, fname) ;
	rf->id = 0 ;

	if ( pos == -1 ) pos = rfcount - 1 ;
	for ( i = pos ; i > 0 ; i-- )
	{
		rfptr[i] = rfptr[i-1] ;
	}
	rfptr[0] = rf ;

	for ( i = 0 ; i < rfcount  ; i++ )
	{
		rfptr[i]->id = IDM_FILESTART + i + 1 ;
		if ( (len = strlen(rfptr[i]->FullPath)) <= MAX_DISPLAY_SIZE - 5)
			sprintf(rfptr[i]->DisplayName, "%d. %s", i+1, rfptr[i]->FullPath) ;
		else
		{
			p = rfptr[i]->FullPath + len - MAX_DISPLAY_SIZE + 7 ;
			sprintf(rfptr[i]->DisplayName, "%d. ...%s", i+1, p) ;
		}
		AppendMenu(	GetSubMenu(GetMenu(hwMDIFrame), FILEMENU),
					MF_STRING|MF_UNCHECKED|MF_ENABLED,
					rfptr[i]->id, rfptr[i]->DisplayName) ;
	}
	DrawMenuBar(hwMDIFrame) ;

}


/*
 * Creates a new EDWIN structure, links it into the list
 * and creates the MDI child window.
 */
EDWIN *NewEditorWindow(void)
{
	EDWIN *ed = new EDWIN ;
	unsigned long lasterr ;

	ed->app = NULL ;
	ed->hWnd = NULL ;
	ed->next = NULL ;
	ed->prev = NULL ;

	if ( Top == NULL )
	{
		EnableMenu() ;
		Top = ed ;
	}
	else
	{
		ed->prev = Last ;
		Last->next = ed ;
	}
	Last = Active = ed ;

	ed->app = new ScrEApp ;
	if ( ed->app == NULL )
	{
		MessageBox(0,"Failed to create new Editor","ScriptEd - Error", MB_OK) ;
		return NULL ;
	}

	ed->app->hInstance = hInst ;
	ed->app->fnbuff[0] = 0 ;
	if ( fnbuff[0] != 0 ) strcpy( ed->app->fnbuff, (const char *)fnbuff) ;
	fnbuff[0] = 0 ;
	ed->app->wMain = ::CreateWindowEx(
	            WS_EX_CLIENTEDGE|WS_EX_MDICHILD,
		        className,
			    "EditorWindow",
				WS_CAPTION | WS_SYSMENU | WS_THICKFRAME |
			    WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_CLIPCHILDREN,
				CW_USEDEFAULT, CW_USEDEFAULT, 
				CW_USEDEFAULT, CW_USEDEFAULT, 
				hwMDIClient,
				NULL,
				hInst,
				0);

	if ( ed->app->wMain == NULL )
	{
		lasterr = GetLastError() ;
		MessageBox(0,"Failed to create MDI child window","ScriptEd",MB_OK) ;
		return NULL ;
	}

	ed->hWnd = ed->app->wMain ;
	ed->app->SetTitle();
	return ed ;
}


/*
 * Find an editor window from it's HWND.
 * Needed all over the place.
 */
EDWIN *FindEditor(HWND hWnd)
{
	EDWIN *ed = Top ;

	if ( Top != NULL )
	{
		do
		{
			if ( ed->hWnd == hWnd ) return ed ;
			ed = ed->next ;
		} while (ed != NULL ) ;
	}
	return NULL ;
}


/*
 * Removes an EDWIN structure from our list and destroys
 * the MDI child window.
 */
void RemoveEditorWindow(HWND hWnd)
{
	EDWIN *ed = FindEditor(hWnd) ;


	if ( ed != NULL )
	{
		if ( ed->next != NULL )
		{
			if ( ed->prev != NULL )
				ed->prev->next = ed->next ;
			ed->next->prev = ed->prev ;
		}
		else
		{
			if ( ed->prev != NULL )
				ed->prev->next = NULL ;
		}
		if ( ed == Top )
		{
			if ( ed->next == NULL )
			{
				Top = NULL ;
				Last = NULL ;
				Active = NULL ;
			}
			else
			{
				Top = ed->next ;
				Top->prev = NULL ;
				Active = ed->next ;
			}
		}
		else
		{
			Active = ed->prev ;
			if ( Last == ed ) Last = Active ;
		}

		DestroyWindow(ed->app->wEditor);
		// Use this message rather than DestroyWindow() otherwise
		// Window-list is not maintained correctly
		SendMessage(hwMDIClient, WM_MDIDESTROY, (WPARAM)ed->hWnd, 0) ;
		delete ed->app ;
		delete ed ;
	}
}


/*
 * Find the current active editor window by querying the
 * MDI client for the active window's HWND.
 */
EDWIN *GetActiveEditor(void)
{
	EDWIN *ed ;

	ed = FindEditor((HWND)SendMessage(hwMDIClient, WM_MDIGETACTIVE, 0, NULL)) ;
	if ( ed != NULL ) Active = ed ;
	if ( Top == NULL )
		EnableMenuItem(::GetMenu(hwMDIFrame), IDM_FILE_PROPERTIES, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	else
		EnableMenuItem(::GetMenu(hwMDIFrame), IDM_FILE_PROPERTIES, MF_ENABLED | MF_BYCOMMAND);
	return ed ;
}


void RemoveDefconsts(void)
{
	ACWORD *acw = ACWTop, *tmp ;


	do
	{
		if ( acw->type == 'U' )
		{
			tmp = acw ;
			if ( acw == ACWTop )
			{
				acw = acw->next ;
				acw->prev = NULL ;
				ACWTop = acw ;
			}
			else if (acw == ACWLast )
			{
				ACWLast = acw->prev ;
				ACWLast->next = NULL ;
				acw = NULL ;
			}
			else
			{
				acw->prev->next = acw->next ;
				acw->next->prev = acw->prev ;
				acw = acw->next ;
			}
			delete tmp ;
		}
		else acw = acw->next ;
	} while ( acw != NULL ) ;

	UpdateACWords() ;
}


void RescanDefconsts(void)
{
	EDWIN *ed = Top ;

	if ( Top != NULL )
	{
		do
		{
			ScanFileDefconsts(ed) ;
			ed = ed->next ;
		} while (ed != NULL ) ;
	}
}


void ScanFileDefconsts(EDWIN *ed)
{

	TextToFind t ;
	long curpos, bytes, len ;
	int line, ukwc = 0 ;
	char *p, *q ;


	curpos = ed->app->SendEditor(SCI_GETCURRENTPOS) ;
	bytes = ed->app->SendEditor(SCI_GETLENGTH) ;

	t.chrg.cpMin = 0 ;
	t.chrg.cpMax = bytes ;
	t.lpstrText = "defconst" ;
	while (ed->app->SendEditor(SCI_FINDTEXT, SCFIND_REGEXP, (long)&t) != -1 )
	{
		line = ed->app->SendEditor(SCI_LINEFROMPOSITION, t.chrgText.cpMin) ;
		len = ed->app->SendEditor(SCI_GETLINE, line, (long)work) ;
		work[len] = 0 ;
		p = work ;
		while (*p)
		{
			switch( *p )
			{
				case ';':
					*p = 0 ;
					break ;

				case '(':
					p++ ; 
					while ( *p == 32 || *p == 9 ) p++ ;
					if ( strncmp((const char *)p, "defconst", 8) == 0)
					{
						p += 8 ;
						while ( *p == 32 || *p == 9 ) p++ ;
						q = p + 1 ;
						while ( *q != 32 && *q != 9 ) q++ ;
						*q = 0 ;
						if ( AddACWord((const char *)p, false) == true )
						{
							UserKWCount++ ;
							ukwc++ ;
						}
						*p = 0 ;
					}
					break ;

				default:
					p++ ;
					break ;
			}
		}	
		t.chrg.cpMin = t.chrgText.cpMin + 1 ;
	}

	if ( ukwc > 0 ) UpdateACWords() ;

}


void UpdateACWords(void)
{
	ACWORD *acw ;


	acwords.Reset() ;
	usrwords.Reset() ;
	delete tmpacw ;
	delete acWords ;
	if ( UserKeyWords != NULL ) delete UserKeyWords ;
	acw = ACWTop ;
	do
	{
		if ( acw->type == 'U' )
		{
			usrwords.Append((const char *)acw->word) ;
			usrwords.Append(" ") ;
		}
		acwords.Append((const char *)acw->word) ;
		acwords.Append(" ") ;
		acw = acw->next ;
	} while ( acw != NULL ) ;
	acWords = acwords.ToString() ;
	UserKeyWords = usrwords.ToString() ;
	tmpacw = new char [strlen((const char *)acWords)+1] ;
	ApplySettings() ;
}


void GetFileStats(void)
{
	long curpos, bytes, rulecount, constcount, lines ;
	TextToFind t ;

	curpos = Active->app->SendEditor(SCI_GETCURRENTPOS) ;
	bytes = Active->app->SendEditor(SCI_GETLENGTH) ;
	lines = Active->app->SendEditor(SCI_GETLINECOUNT) ;
	rulecount = constcount = 0 ;
	t.chrg.cpMin = 0 ;
	t.chrg.cpMax = bytes ;
	t.lpstrText = "defrule" ;
	while (Active->app->SendEditor(SCI_FINDTEXT, SCFIND_REGEXP, (long)&t) != -1 )
	{
		rulecount++ ;
		t.chrg.cpMin = t.chrgText.cpMin + 1 ;
	}

	t.chrg.cpMin = 0 ;
	t.chrg.cpMax = bytes ;
	t.lpstrText = "defconst" ;
	while (Active->app->SendEditor(SCI_FINDTEXT, SCFIND_REGEXP, (long)&t) != -1 )
	{
		constcount++ ;
		t.chrg.cpMin = t.chrgText.cpMin + 1 ;
	}
	sprintf(work, "Lines:\t%ld\r\nBytes:\t%ld\r\nRules:\t%ld\r\nConsts:\t%ld\r\n",
			lines, bytes, rulecount, constcount) ;

	MessageBox(hwMDIFrame, work, "Properties", MB_OK) ;
}


/*
 * Search for text within the document. We use regular expressions.
 * Triggered from the Find and Find+Replace dialogues and F3 keypress.
 */
int FindSearchText(int next)
{

	int r = -1 ;

      if ( fText[0] != 0 )
      {
        ttf.chrg.cpMin = Active->app->SendEditor(SCI_GETCURRENTPOS) + next;
        ttf.chrg.cpMax = Active->app->SendEditor(SCI_GETLENGTH) ;
        ttf.lpstrText = fText ;
        if ( (r = Active->app->SendEditor(SCI_FINDTEXT, SCFIND_REGEXP, (long)&ttf)) != -1 )
        {
          Active->app->SendEditor(SCI_GOTOPOS, ttf.chrgText.cpMin) ;
          Active->app->SendEditor(SCI_SETSELECTIONSTART, ttf.chrgText.cpMin) ;
          Active->app->SendEditor(SCI_SETSELECTIONEND, ttf.chrgText.cpMax) ;
        }
		else
		{
			Active->app->SendEditor(SCI_GOTOPOS, 0) ;
	        ttf.chrg.cpMin = 0 ;
	        if ( (r = Active->app->SendEditor(SCI_FINDTEXT, SCFIND_REGEXP, (long)&ttf)) != -1 )
			{
				Active->app->SendEditor(SCI_GOTOPOS, ttf.chrgText.cpMin) ;
				Active->app->SendEditor(SCI_SETSELECTIONSTART, ttf.chrgText.cpMin) ;
				Active->app->SendEditor(SCI_SETSELECTIONEND, ttf.chrgText.cpMax) ;
			}
			else MessageBox(hwMDIFrame, "Text not found","Find", MB_OK) ;
		}
      }
      Active->app->SendEditor(SCI_SETFOCUS, true) ;

  return r ;

}


/*
 * Initialise the Goto Line dialogue items. 
 * Defaults to current line number.
 */
void InitGotoDlgOptions(HWND hDlg)
{
	char dlgitem[MAX_PATH] ;
	long pos ;

	dlgFocus = false ;
	dlgCancel = false ;
	pos = Active->app->SendEditor(SCI_GETCURRENTPOS) ;
	LineToGoto = -1 ;
	sprintf(dlgitem, "%ld", Active->app->SendEditor(SCI_LINEFROMPOSITION,pos)+1) ;
	SetDlgItemText(hDlg, IDC_EDIT_LNUM, (const char *)dlgitem) ;
	SendMessage(GetDlgItem(hDlg, IDC_EDIT_LNUM), EM_SETSEL, 0, -1) ;
}


/*
 * Initialise Find dialogue items.
 * Defaults to previous search text.
 */
void InitFindDlgOptions(HWND hDlg)
{
	dlgFocus = false ;
	dlgCancel = false ;
	SetDlgItemText(hDlg, IDC_EDIT_FIND, (const char *)fText) ;
}


/*
 * Initialise Find+Replace dialogue items.
 * Defaults to previous search and replace texts.
 */
void InitReplaceDlgOptions(HWND hDlg)
{
	dlgFocus = false ;
	dlgCancel = false ;
	SetDlgItemText(hDlg, IDC_EDIT_FINDWHAT, (const char *)fText) ;
	SendMessage(GetDlgItem(hDlg, IDC_EDIT_FINDWHAT), EM_SETSEL, 0, -1) ;
	SetDlgItemText(hDlg, IDC_EDIT_REPLACEWITH, (const char *)rText) ;
}


/*
 * Callback function for enumerating available fonts.
 * The MSDN information for this is incorrect with regards
 * the types of parameters :(
 */
int __stdcall ScreenFontEnumProc(	const LOGFONT *lpelfe, const TEXTMETRIC *lpntme,
									DWORD FontType, LPARAM lParam)
{
	HWND hw = (HWND)lParam ;
	LRESULT l ;


	// Need to do this check because the same font-family name
	// can be returned more than once.
	if ( SendDlgItemMessage(hw, IDC_SET_COMBO_FONT1, CB_FINDSTRINGEXACT, 0, (LPARAM)lpelfe->lfFaceName) == CB_ERR )
	{
		if ( lpelfe->lfFaceName[0] != '@' )
		{ 
			l = SendDlgItemMessage(hw, IDC_SET_COMBO_FONT1, CB_ADDSTRING, 0, (LPARAM)lpelfe->lfFaceName); 
			if ( l == CB_ERR ) return 0 ;
			l = SendDlgItemMessage(hw, IDC_SET_COMBO_FONT2, CB_ADDSTRING, 0, (LPARAM)lpelfe->lfFaceName); 
			if ( l == CB_ERR ) return 0 ;
		}
	}
	return 1 ; // returning zero terminates the enumeration.
}


/*
 * Initialise Settings dialogue items
 */
void InitSettingsDlgOptions(HWND hDlg)
{
	HDC hdc ;
	LOGFONT lf ;
	int i ;


	dlgFocus = false ;
	dlgCancel = false ;
	// Fill font combo-boxes with available fonts
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_FONT1, CB_RESETCONTENT, 0, 0); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_FONT2, CB_RESETCONTENT, 0, 0); 
	// Get device context - needed for font enumeration
//	if ( Active != NULL )
//		hdc = GetDC(Active->app->wMain) ;
//	else
		hdc = GetDC(hwMDIFrame) ;
	lf.lfCharSet = DEFAULT_CHARSET ;
	lf.lfFaceName[0] = 0 ;
	lf.lfPitchAndFamily = 0 ;
	EnumFontFamiliesEx(hdc, &lf, ScreenFontEnumProc, (LPARAM)hDlg,0) ;
//	ReleaseDC(Active->app->wMain, hdc) ;
	ReleaseDC(hwMDIFrame, hdc) ;

	// Fill font-size combo-boxes
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_RESETCONTENT, 0, 0); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_RESETCONTENT, 0, 0); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_ADDSTRING, 0, (LPARAM)"6"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_ADDSTRING, 0, (LPARAM)"6"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_ADDSTRING, 0, (LPARAM)"8"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_ADDSTRING, 0, (LPARAM)"8"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_ADDSTRING, 0, (LPARAM)"10"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_ADDSTRING, 0, (LPARAM)"10"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_ADDSTRING, 0, (LPARAM)"12"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_ADDSTRING, 0, (LPARAM)"12"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_ADDSTRING, 0, (LPARAM)"14"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_ADDSTRING, 0, (LPARAM)"14"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_ADDSTRING, 0, (LPARAM)"16"); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_ADDSTRING, 0, (LPARAM)"16"); 


	// set current font size settings
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_SETCURSEL, (comFontSize/2) - 3, 0); 
	SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_SETCURSEL, (cmdFontSize/2) - 3, 0); 

	// Set current font face settings
	i = SendDlgItemMessage(hDlg, IDC_SET_COMBO_FONT1, CB_FINDSTRINGEXACT, 0, (LPARAM)comFont) ;
    if ( i != CB_ERR )
		SendDlgItemMessage(hDlg, IDC_SET_COMBO_FONT1, CB_SETCURSEL, i, 0); 
	i = SendDlgItemMessage(hDlg, IDC_SET_COMBO_FONT2, CB_FINDSTRINGEXACT, 0, (LPARAM)cmdFont) ;
    if ( i != CB_ERR )
		SendDlgItemMessage(hDlg, IDC_SET_COMBO_FONT2, CB_SETCURSEL, i, 0); 
	
	// Set auto-complete option
	SendDlgItemMessage(hDlg, IDC_SET_ALLOW_AUTOC, BM_SETCHECK, AllowAutoC, 0) ;

	// set current colours
	pbinf[0].hWnd = GetDlgItem(hDlg, IDC_CC_PB1) ;
	pbinf[1].hWnd = GetDlgItem(hDlg, IDC_CC_PB2) ;
	pbinf[2].hWnd = GetDlgItem(hDlg, IDC_CC_PB3) ;
	pbinf[3].hWnd = GetDlgItem(hDlg, IDC_CC_PB4) ;
	pbinf[4].hWnd = GetDlgItem(hDlg, IDC_CC_PB5) ;
	pbinf[5].hWnd = GetDlgItem(hDlg, IDC_CC_PB6) ;
	pbinf[6].hWnd = GetDlgItem(hDlg, IDC_CC_PB7) ;
	pbinf[7].hWnd = GetDlgItem(hDlg, IDC_CC_PB8) ;
	pbinf[8].hWnd = GetDlgItem(hDlg, IDC_CC_PB9) ;
	pbinf[9].hWnd = GetDlgItem(hDlg, IDC_CC_PB10) ;
	pbinf[10].hWnd = GetDlgItem(hDlg, IDC_CC_PB11) ;
	pbinf[0].cref = bgColour ;
	pbinf[1].cref = fgColour ;
	pbinf[2].cref = kwColour ;
	pbinf[3].cref = paColour ;
	pbinf[4].cref = acColour ;
	pbinf[5].cref = faColour ;
	pbinf[6].cref = ukColour ;
	pbinf[7].cref = syColour ;
	pbinf[8].cref = coColour ;
	pbinf[9].cref = nuColour ;
	pbinf[10].cref = qsColour ;

}

/*
 * Get the Settings dialogue items when the dialogue closes
 */
void GetSettingsDlgOptions(HWND hDlg)
{
	comFontSize = SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE1, CB_GETCURSEL, 0, 0) * 2 + 6; 
	cmdFontSize = SendDlgItemMessage(hDlg, IDC_SET_COMBO_SIZE2, CB_GETCURSEL, 0, 0) * 2 + 6; 
	GetDlgItemText(hDlg, IDC_SET_COMBO_FONT1, comFont, FONT_NAME_SIZE) ;
	GetDlgItemText(hDlg, IDC_SET_COMBO_FONT2, cmdFont, FONT_NAME_SIZE) ;
	AllowAutoC = SendDlgItemMessage(hDlg, IDC_SET_ALLOW_AUTOC, BM_GETCHECK, 0, 0); 

	bgColour = pbinf[0].cref ;
	fgColour = pbinf[1].cref ;
	kwColour = pbinf[2].cref ;
	paColour = pbinf[3].cref ;
	acColour = pbinf[4].cref ;
	faColour = pbinf[5].cref ;
	ukColour = pbinf[6].cref ;
	syColour = pbinf[7].cref ;
	coColour = pbinf[8].cref ;
	nuColour = pbinf[9].cref ;
	qsColour = pbinf[10].cref ;

}


/*
 * Get the line number to go to from the Goto Line dialogue
 */
void GetGotoDlgOptions(HWND hDlg)
{
	char dlgitem[MAX_PATH] ;


	GetDlgItemText(hDlg, IDC_EDIT_LNUM, dlgitem, MAX_PATH-1) ;
	LineToGoto = atoi((const char *)dlgitem) ;
	LineToGoto = ( LineToGoto > 0 ) ? LineToGoto - 1 : 0 ;
 }


/*
 * Get the text to search for from Find dialogue
 */
void GetFindDlgOptions(HWND hDlg)
{
	GetDlgItemText(hDlg, IDC_EDIT_FIND, fText, MAX_PATH-1) ; 
}


/*
 * Get the search and replace texts from the Replace dialogue
 */
void GetReplaceDlgOptions(HWND hDlg)
{
	GetDlgItemText(hDlg, IDC_EDIT_FINDWHAT, fText, MAX_PATH-1) ; 
	GetDlgItemText(hDlg, IDC_EDIT_REPLACEWITH, rText, MAX_PATH-1) ; 
}


/*
 * Window procedure for the Goto Line dialogue.
 */
int __stdcall GotoWndProc(	HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
			hwCurrentDialogue = hDlg ;
			InitGotoDlgOptions(hDlg) ;
			SetFocus(GetDlgItem(hDlg, IDC_EDIT_LNUM)) ;
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_BUT_GOTO_OK:
					GetGotoDlgOptions(hDlg) ;
  					EndDialog(hDlg, LOWORD(wParam));
					hwCurrentDialogue = NULL ;
	  				return TRUE;

				case IDC_BUT_GOTO_CANCEL:
				case IDCANCEL:
					EndDialog(hDlg, LOWORD(wParam));
					hwCurrentDialogue = NULL ;
					return TRUE;
			}
			break ;
	}
	return FALSE;
}


/*
 * Window procedure for the About dialogue.
 */
int __stdcall AboutWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
			hwCurrentDialogue = hDlg ;
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_BUT_CREDITS:
					MessageBox(hDlg,credits,"Credits",MB_OK) ;
					return TRUE;

				case IDCANCEL:
				case IDC_ABOUT_CLOSE:
  					EndDialog(hDlg, LOWORD(wParam));
					hwCurrentDialogue = NULL ;
	  				return TRUE;
			}
			break ;
	}
	return FALSE;
}


int __stdcall RefWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	int cat, subcat ;
	char kwname[MAX_PATH] ;
	const char *tip_usage, *tip_desc ;
	char *tmptd, *p ;

	switch (message)
	{
		case WM_INITDIALOG:
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_COMBO_CAT:
				case IDC_COMBO_SUBCAT:
					cat = SendDlgItemMessage(hDlg, IDC_COMBO_CAT, CB_GETCURSEL, 0, 0); 
					subcat = SendDlgItemMessage(hDlg, IDC_COMBO_SUBCAT, CB_GETCURSEL, 0, 0); 
					if ( cat != LastCat || subcat != LastSubCat )
					{
						SetKeywordsCombo(cat,subcat) ;
						LastCat = cat ;
						LastSubCat = subcat ;
					}
					break ;

				case IDC_REF_SHOW:
					GetDlgItemText(hDlg, IDC_COMBO_KW, kwname, MAX_PATH) ;
					if (ScriptXML.FindNode("keyword", "name", (const char *)kwname) == 0 )
					{
						ScriptXML.MoveToFirstChild() ;
						tip_usage = ScriptXML.NodeValue() ;
						ScriptXML.MoveToNextSibling() ;
						tip_desc = ScriptXML.NodeValue() ;
						tmptd = new char [strlen(tip_desc)+1] ;
						strcpy(tmptd, tip_desc) ;
						p = tmptd ;
						while (*p) 
						{
							if ( *p == '\\' && *(p+1) == 'n')
							{
								*p = 13 ;
								*(p+1) = 10 ;
							}
							p++ ;
						}
						if ( strcmp((const char *)RefCat[LastCat], "ARTICLE") == 0 )
							sprintf(RefDesc, "%s",tmptd) ;
						else
							sprintf(RefDesc, "Usage: %s\r\n\r\nDescription:\r\n%s", tip_usage, tmptd) ;
						SetDlgItemText(hDlg, IDC_REF_INFO, (const char *)RefDesc) ;
					}
					break ;

				case IDCANCEL:
				case IDC_REF_CLOSE:
  					DestroyWindow(hwRefDialogue);
					hwRefDialogue = NULL ;
					LastCat = 0 ;
					LastSubCat = 0 ;
	  				return TRUE;
			}
			break ;
	}
	return FALSE;

}


/*
 * Window procedure for the Find dialogue
 */
int __stdcall FindWndProc(	HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{


	switch (message)
	{
		case WM_INITDIALOG:
			hwCurrentDialogue = hDlg ;
	        InitFindDlgOptions(hDlg) ;
			return TRUE;

		case WM_PAINT:
			if ( dlgFocus == false )
			{
				SetFocus(GetDlgItem(hDlg, IDC_EDIT_FIND)) ;
				if ( fText[0] != 0)
					SendMessage(GetDlgItem(hDlg, IDC_EDIT_FIND), EM_SETSEL, 0, -1) ;
				dlgFocus = true ;
			}
			break ;


		case WM_COMMAND:
          switch(LOWORD(wParam))
          {
						
			  case IDC_FIND_OK:
                GetFindDlgOptions(hDlg) ;
  				EndDialog(hDlg, LOWORD(wParam));
				hwCurrentDialogue = NULL ;
	  			return TRUE;
                break ;

				case IDCANCEL:
              case IDC_FIND_CANCEL:
  				EndDialog(hDlg, LOWORD(wParam));
				hwCurrentDialogue = NULL ;
				dlgCancel = true ;
	  			return TRUE;
    			break;
		  }
          break ;
	}
  return FALSE;

}


/*
 * Re-initialise all the editor windows with the new
 * colours and fonts
 */
void ApplySettings(void)
{
	EDWIN *ed = Top ;


	if ( Top != NULL )
	{
		do
		{
			ed->app->InitialiseEditor() ;
			ed = ed->next ;
		} while ( ed != NULL ) ;
	}
}


/*
 * Close all open MDI child windows. Checks to see whether
 * the documents have changed.
 */
bool CloseAllWindows(void)
{
	EDWIN *ed = Top, *n ;


	if ( Top != NULL )
	{
		do
		{
			if (ed->app->SaveIfUnsure() != IDCANCEL)
			{
				n = ed->next ;
				RemoveEditorWindow(ed->hWnd) ;
				ed = n ;
			}
			else return false ;
		} while ( ed != NULL ) ;
	}
	return true ;
}


/*
 * Restores all maximised or minimised MDI child windows
 * to their standard sizes.
 */
void RestoreChildren(void)
{
	EDWIN *ed = Top ;


	if ( Top != NULL )
	{
		do
		{
			SendMessage(hwMDIClient, WM_MDIRESTORE, (WPARAM)ed->hWnd, 0) ;
			ed = ed->next ;
		} while ( ed != NULL ) ;
	}
}


void DoColourPicker(HWND hwnd, int butId)
{
	int pbId, i ;
	COLORREF cref ;
	unsigned short r,g,b,a ;
	SColour colour ;

	switch( butId )
	{
		case IDC_CCP1:
			pbId = IDC_CC_PB1 ;
			cref = bgColour ;
			i = 0 ;
			break ;

		case IDC_CCP2:
			pbId = IDC_CC_PB2 ;
			cref = fgColour ;
			i = 1 ;
			break ;

		case IDC_CCP3:
			pbId = IDC_CC_PB3 ;
			cref = kwColour ;
			i = 2 ;
			break ;

		case IDC_CCP4:
			pbId = IDC_CC_PB4 ;
			cref = paColour ;
			i = 3 ;
			break ;

		case IDC_CCP5:
			pbId = IDC_CC_PB5 ;
			cref = acColour ;
			i = 4 ;
			break ;

		case IDC_CCP6:
			pbId = IDC_CC_PB6 ;
			cref = faColour ;
			i = 5 ;
			break ;

		case IDC_CCP7:
			pbId = IDC_CC_PB7 ;
			cref = ukColour ;
			i = 6 ;
			break ;

		case IDC_CCP8:
			pbId = IDC_CC_PB8 ;
			cref = syColour ;
			i = 7 ;
			break ;

		case IDC_CCP9:
			pbId = IDC_CC_PB9 ;
			cref = coColour ;
			i = 8 ;
			break ;

		case IDC_CCP10:
			pbId = IDC_CC_PB10 ;
			cref = nuColour ;
			i = 9 ;
			break ;

		case IDC_CCP11:
			pbId = IDC_CC_PB11 ;
			cref = qsColour ;
			i = 10 ;
			break ;

	}

	r = GetRValue(cref) ;
	g = GetGValue(cref) ;
	b = GetBValue(cref) ;
	a = 100 ;
	cpicker = new CColourPicker(hwnd, r, g, b, a, true);
	cpicker->CreatecolourPicker(CP_NO_ALPHA);
	SendDlgItemMessage(hwnd, pbId, WM_PAINT, 0, 0);	
	colour = cpicker->GetCurrentColour() ;
	pbinf[i].cref = RGB(colour.r,colour.g,colour.b) ;
	delete cpicker ;
	cpicker = NULL ;

}


int GetPBControlID(HWND hWnd)
{
	int i ;

	for ( i = 0 ; i < MAX_COLOURS ; i++ )
		if ( pbinf[i].hWnd == hWnd ) return pbinf[i].id ;
	return 0 ;
}


void SetDefaultPBColour(int id, SColour *colour)
{
	COLORREF cref ;

	colour->a = 100 ;
	switch(id)
	{
		case IDC_CC_PB1:
			cref = bgColour ;
			break ;
		case IDC_CC_PB2:
			cref = fgColour ;
			break ;
		case IDC_CC_PB3:
			cref = kwColour ;
			break ;
		case IDC_CC_PB4:
			cref = paColour ;
			break ;
		case IDC_CC_PB5:
			cref = acColour ;
			break ;
		case IDC_CC_PB6:
			cref = faColour ;
			break ;
		case IDC_CC_PB7:
			cref = ukColour ;
			break ;
		case IDC_CC_PB8:
			cref = syColour ;
			break ;
		case IDC_CC_PB9:
			cref = coColour ;
			break ;
		case IDC_CC_PB10:
			cref = nuColour ;
			break ;
		case IDC_CC_PB11:
			cref = qsColour ;
			break ;
		default:
			cref = 0 ;
			break ;
	}
	colour->r = GetRValue(cref) ;
	colour->g = GetGValue(cref) ;
	colour->b = GetBValue(cref) ;
}


/*
 * Window procedure for Settings dialogue  - ID = 106
 */
int __stdcall SettingsWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{


	switch (message)
	{
		case WM_INITDIALOG:
			hwCurrentDialogue = hDlg ;
			InitSettingsDlgOptions(hDlg) ;
			return TRUE;

		case WM_PAINT:
          //SetFocus(GetDlgItem(hDlg, IDC_SET_COMBO_FONT1)) ;
          break ;

		case WM_CTLCOLORSTATIC:
			{
				SColour colour ;
				if ( cpicker != NULL )
					colour = cpicker->GetCurrentColour();
				else
					SetDefaultPBColour(GetPBControlID((HWND)lParam), &colour) ;
				DrawCheckedRect((HWND)lParam, colour.r, colour.g, colour.b, colour.a, 10, 10);
			}
			return TRUE ;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_CCP1:
				case IDC_CCP2:
				case IDC_CCP3:
				case IDC_CCP4:
				case IDC_CCP5:
				case IDC_CCP6:
				case IDC_CCP7:
				case IDC_CCP8:
				case IDC_CCP9:
				case IDC_CCP10:
				case IDC_CCP11:
					DoColourPicker(hDlg, LOWORD(wParam)) ;
					return TRUE ;
					break ;

				case IDC_SET_APPLY:
					GetSettingsDlgOptions(hDlg) ;
					SetRegistrySettings() ;
					ApplySettings() ;
					return TRUE;
					break ;

				case IDC_FORM_SET_OK:
					GetSettingsDlgOptions(hDlg) ;
					SetRegistrySettings() ;
  					EndDialog(hDlg, LOWORD(wParam));
					hwCurrentDialogue = NULL ;
					ApplySettings() ;
	  				return TRUE;
					break ;

				case IDCANCEL:
				case IDC_FORM_SET_CANCEL:
  					EndDialog(hDlg, LOWORD(wParam));
					hwCurrentDialogue = NULL ;
	  				return TRUE;
    				break;
		  }
          break ;
	}
  return FALSE;

}


/*
 * Window procedure for Find+Replace dialogue
 */
int __stdcall ReplaceWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{

	int r ;

	switch (message)
	{
		case WM_INITDIALOG:
			hwCurrentDialogue = hDlg ;
          InitReplaceDlgOptions(hDlg) ;
  		  return TRUE;

		case WM_PAINT:
          SetFocus(GetDlgItem(hDlg, IDC_EDIT_FINDWHAT)) ;
          break ;

		case WM_COMMAND:
          GetReplaceDlgOptions(hDlg) ;
          switch(LOWORD(wParam))
          {
			  case IDC_REPLACE_ALL:
                r = FindSearchText(0) ;
                while ( r != -1 )
                {
                  Active->app->SendEditor(SCI_SETTARGETSTART, ttf.chrgText.cpMin) ;
                  Active->app->SendEditor(SCI_SETTARGETEND, ttf.chrgText.cpMax) ;
                  Active->app->SendEditor(SCI_REPLACETARGET, -1, (long)rText) ;
                  r = FindSearchText(1) ;
                }
                break ;

			  case IDC_REPLACE:
                Active->app->SendEditor(SCI_SETTARGETSTART, ttf.chrgText.cpMin) ;
                Active->app->SendEditor(SCI_SETTARGETEND, ttf.chrgText.cpMax) ;
                Active->app->SendEditor(SCI_REPLACETARGET, -1, (long)rText) ;
                FindSearchText(1) ;
                break ;

			  case IDC_REPLACE_FIND_NEXT:
                FindSearchText(1) ;
	  			return TRUE;
                break ;

				case IDCANCEL:
              case IDC_REPLACE_CLOSE:
  				EndDialog(hDlg, LOWORD(wParam));
				hwCurrentDialogue = NULL ;
				dlgCancel = true ;
	  			return TRUE;
    			break;
		  }
          break ;
	}
  return FALSE;

}


void OpenRecentFile(unsigned int id)
{
	EDWIN *ed ;
	int i ;


	for (i = 0 ; i < rfcount ; i++)
	{
		if ( Recent[i].id == id )
			break ;
	}

	if ( (ed = AlreadyOpen(Recent[i].FullPath)) == NULL )
	{
		NewEditorWindow() ;
		GetActiveEditor() ;
		Active->app->OpenFile(Recent[i].FullPath);
		UpdateFileMenu(Recent[i].FullPath, Active) ;
		ScanFileDefconsts(Active) ;

	}
	else
		PostMessage(hwMDIClient, WM_MDIACTIVATE, (WPARAM)ed->hWnd, 0) ;
}


EDWIN *AlreadyOpen(const char *fname)
{
	EDWIN *ed = Top ;


	if ( Top != NULL )
	{
		do
		{
			if ( strcmp(fname, (const char *)ed->app->fullPath) ==0 )
				return ed ;
			ed = ed->next ;
		} while ( ed != NULL ) ;
	}
	return NULL ;
}


/*
 * Window procedure for the MDI Frame window.
 * Creates the MDI Client window when it receives 
 * the WM_CREATE message. Many messages need to be
 * handled by DefFrameProc() regardless of whether
 * we have processed the message.
 */
LRESULT CALLBACK MDIFrameWndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	char openName[MAX_PATH] = "\0";
	char *filter = "AoC Scripts (.per)\0*.per;\0All Files (*.*)\0*.*\0\0";
	OPENFILENAME ofn = {sizeof(OPENFILENAME)};
	EDWIN *ed ;


	switch (iMessage)
	{
		case WM_CREATE:

			CLIENTCREATESTRUCT ccs; 
 
	        // Get menu handle and assign the first child window identifier. 
 			ccs.hWindowMenu = GetSubMenu(GetMenu(hWnd), WINDOWMENU); 
			ccs.idFirstChild = IDM_WINDOWCHILD; 


	       // Create the MDI client window. 
	        hwMDIClient = CreateWindowEx(0,
										 "MDICLIENT", 
										"MDI Client Window", 
								        WS_CHILD | WS_CLIPCHILDREN | 
										WS_VSCROLL | WS_HSCROLL | WS_VISIBLE, 
										0, 0, 0, 0, 
										hWnd, 
										(HMENU) 0xCAC,  // was NULL
										hInst, (char *)&ccs); 
	        ShowWindow(hwMDIClient, SW_SHOW); 
			InitToolTips(hWnd) ;
		 	InitToolBar(hWnd) ;
			InitStatusBar(hWnd) ;
//			return 0 ;
			break ;

		case WM_PAINT:
			break ;

		case WM_COMMAND:			// Need to pass this to DefFrameProc too
			GetActiveEditor() ;
			switch (LOWORD(wParam))
			{
				case IDM_CLEAR_DEFCONSTS:
					RemoveDefconsts() ;
					break ;

				case IDM_RESCAN_DEFCONSTS:
					RescanDefconsts() ;
					break ;

				case IDM_FILE_PROPERTIES:
					if (GetActiveEditor() != NULL )	GetFileStats() ;
					break ;

				case IDM_RF1:
				case IDM_RF2:
				case IDM_RF3:
				case IDM_RF4:
				case IDM_RF5:
				case IDM_RF6:
				case IDM_RF7:
				case IDM_RF8:
				case IDM_RF9:
				case IDM_RF10:
					OpenRecentFile(LOWORD(wParam)) ;
					break ;

				case IDM_HELP_REF:
					//if (hwRefDialogue == NULL) InitialiseRef() ;
					if ( hwHelp == NULL ) 
						CreateHelpWindow() ;
					break ;

				case IDM_HELP_ABOUT:
					DialogBox(hInst, MAKEINTRESOURCE(IDD_DLG_ABOUT),hwMDIFrame, AboutWndProc) ;
					break ;

				case IDM_TIPS:
					if ( TipsSwitchedOn == true )
					{
						SendMessage(hwStatusBar, SB_SETTEXT, 2, (LPARAM)"Tips OFF") ;
						TipsSwitchedOn = false ;
						CheckMenuItem(GetSubMenu(GetMenu(hWnd), TOOLSMENU), 0, MF_BYPOSITION|MF_UNCHECKED) ;
					}
					else
					{
						SendMessage(hwStatusBar, SB_SETTEXT, 2, (LPARAM)"Tips ON") ;
						TipsSwitchedOn = true ;
						CheckMenuItem(GetSubMenu(GetMenu(hWnd), TOOLSMENU), 0, MF_BYPOSITION|MF_CHECKED) ;
					}
					break ;

				case IDM_MDI_NEXT:
					if ( Active != NULL )
						PostMessage(hwMDIClient, WM_MDINEXT, (WPARAM)Active->hWnd, 0) ; 
					break ;

				case IDM_MDI_PREV:
					if ( Active != NULL )
						PostMessage(hwMDIClient, WM_MDINEXT, (WPARAM)Active->hWnd, 1) ; 
					break ;

				case IDM_WIN_TILE_H:
					PostMessage(hwMDIClient, WM_MDITILE, MDITILE_HORIZONTAL, 0) ; 
					break ;

				case IDM_WIN_TILE_V:
					PostMessage(hwMDIClient, WM_MDITILE, MDITILE_VERTICAL, 0) ; 
					break ;

				case IDM_WIN_CASCADE:
					PostMessage(hwMDIClient, WM_MDICASCADE, MDITILE_SKIPDISABLED, 0) ;
					break ;

				case IDM_FILE_EXIT:
					PostMessage(hwMDIFrame, WM_CLOSE, 0, 0) ;
					break ;

				case IDM_FILE_NEW:
					NewEditorWindow() ;
					GetActiveEditor() ;
					Active->app->New() ;				
					break ;

				case IDM_FILE_OPEN:
					RestoreChildren() ;
					ofn.hwndOwner = hwMDIFrame;
					ofn.hInstance = hInst;
					ofn.lpstrFile = openName;
					ofn.nMaxFile = sizeof(openName);
					ofn.lpstrFilter = filter;
					ofn.lpstrCustomFilter = 0;
					ofn.nMaxCustFilter = 0;
					ofn.nFilterIndex = 0;
					ofn.lpstrTitle = "Open File";
					ofn.Flags = OFN_HIDEREADONLY;
					if (::GetOpenFileName(&ofn))
					{
						if ( (ed = AlreadyOpen(openName)) == NULL )
						{
							NewEditorWindow() ;
							GetActiveEditor() ;
							Active->app->OpenFile(openName);
							UpdateFileMenu(openName, Active) ;
							ScanFileDefconsts(Active) ;
						}
						else
							PostMessage(hwMDIClient, WM_MDIACTIVATE, (WPARAM)ed->hWnd, 0) ;
					}
					break ;

				case IDM_FILE_SAVE:
					if ( Active->app->fullPath[0] != 0 )
					{
						Active->app->Save();
						UpdateFileMenu(Active->app->fullPath, Active) ;
						break;
					} // else drop through to "save as"

				case IDM_FILE_SAVEAS:
					strcpy(openName, Active->app->fullPath);
					ofn.hwndOwner = hwMDIFrame;
					ofn.hInstance = hInst;
					ofn.lpstrFile = openName;
					ofn.nMaxFile = sizeof(openName);
					ofn.lpstrTitle = "Save File";
					ofn.Flags = OFN_HIDEREADONLY;
					if (::GetSaveFileName(&ofn))
					{
						strcpy(Active->app->fullPath, openName);
						Active->app->SetTitle();
						Active->app->SaveFile(Active->app->fullPath);
						::InvalidateRect(Active->app->wEditor, 0, 0);
						UpdateFileMenu(Active->app->fullPath, Active) ;
					}
					break;

				case IDM_CLOSE:
					GetActiveEditor() ;
					if (Active->app->SaveIfUnsure() != IDCANCEL)
					{
						RemoveEditorWindow(Active->hWnd) ;
						GetActiveEditor() ;
					}
					break ;

				case IDM_EDIT_UNDO:
					Active->app->SendEditor(WM_UNDO);
					break;

				case IDM_EDIT_REDO:
					Active->app->SendEditor(SCI_REDO);
					break;

				case IDM_EDIT_CUT:
					Active->app->SendEditor(WM_CUT);
					break;

				case IDM_EDIT_COPY:
					Active->app->SendEditor(WM_COPY);
					break;

				case IDM_EDIT_PASTE:
					Active->app->SendEditor(WM_PASTE);
					break;

				case IDM_EDIT_DELETE:
					Active->app->SendEditor(WM_CLEAR);
					break;

				case IDM_EDIT_SELECTALL:
					Active->app->SendEditor(SCI_SELECTALL);
					break;

				case IDM_GOTO_LINE:
					DialogBox(hInst, MAKEINTRESOURCE(IDD_FORM_GOTO),hwMDIFrame, GotoWndProc) ;
					if ( LineToGoto != -1 )
					{
						Active->app->SendEditor(SCI_GOTOLINE, LineToGoto) ;
					}
					Active->app->SendEditor(SCI_SETFOCUS, true) ;
					break ;

				case IDM_FIND:
					DialogBox(hInst, MAKEINTRESOURCE(IDD_FORM_FIND),hwMDIFrame, FindWndProc) ;
					if ( dlgCancel == false ) FindSearchText(0) ;
					break ;

				case IDM_FIND_NEXT:		// F3
					FindSearchText(1) ;
					break ;

				case IDM_REPLACE:
					DialogBox(hInst, MAKEINTRESOURCE(IDD_FORM_REPLACE),hwMDIFrame, ReplaceWndProc) ;
					break ;

				case IDM_SETTINGS:
					DialogBox(hInst, MAKEINTRESOURCE(IDD_FORM_SETTINGS),hwMDIFrame, SettingsWndProc) ;
					break ;

			}
			if ( Active != NULL ) SetFocus(Active->app->wEditor) ;
			break ;


		case WM_NOTIFY:		// Might need to pass these to all MDI children ... ???
/*			if ( hWnd == hwMDIFrame )
				MessageBox(0,"hello","hello",MB_OK) ;
			if (GetActiveEditor() != NULL ) Active->app->Notify(reinterpret_cast<SCNotification *>(lParam));
//			return 0 ;
*/
			break ;

		case WM_MENUSELECT:
			if (GetActiveEditor() != NULL )	Active->app->CheckMenus();
			break ;

		case WM_MENUCHAR:		// Needs to go to DefFrameProc too
			break ;

		case WM_SETFOCUS:		// Needs to go to DefFrameProc too
			gotmdi = true ;
			break ;

		case WM_SIZE:			// Needs to go to DefFrameProc too
			RECT rectClient, rectStatus, rectTool;
			UINT uToolHeight, uStatusHeight , uClientAlreaHeight;
			SendMessage(hwToolBar, TB_AUTOSIZE, 0, 0);
			SendMessage(hwStatusBar, WM_SIZE, 0, 0);
			GetClientRect(hWnd, &rectClient);
			GetWindowRect(hwStatusBar, &rectStatus);
			GetWindowRect(hwToolBar, &rectTool);
			uToolHeight = rectTool.bottom - rectTool.top;
			uStatusHeight = rectStatus.bottom - rectStatus.top;
			uClientAlreaHeight = rectClient.bottom;
			MoveWindow(hwMDIClient, 0, uToolHeight, rectClient.right,
            uClientAlreaHeight - uStatusHeight - uToolHeight, TRUE);
			return 0 ;
			break ;

	case WM_CLOSE:
		if (CloseAllWindows() == true) 
		{
			if (hwRefDialogue != NULL )
				DestroyWindow(hwRefDialogue);
			::PostQuitMessage(0);
		}
		return 0;

	}
	return DefFrameProc(hWnd, hwMDIClient, iMessage, wParam, lParam ) ;
}


/*
 * Window procedure for MDI children.
 * Creates the Scintilla window during the WM_CREATE event.
 * Sends notifications to child window application class instances.
 */
LRESULT CALLBACK WndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	EDWIN *ed ;


	switch (iMessage)
	{
		case WM_CREATE:
			RECT rc;
			GetClientRect(hWnd, &rc);
			Active->app->wEditor = CreateWindow(
						"Scintilla",
						"Source",
						WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_CLIPCHILDREN,
						0, 0, rc.right - rc.left, rc.bottom - rc.top,
						hWnd, 0, Active->app->hInstance, 0);
			Active->app->InitialiseEditor();
			ShowWindow(Active->app->wEditor, SW_SHOW);
			Active->app->CheckMenus() ;
			SetFocus(Active->app->wEditor) ;
//			return 0;
			break ;

		case WM_MDIACTIVATE:
//			if ( GetActiveEditor() != NULL ) Active->app->CheckMenus() ;
			if ( (ed = FindEditor((HWND)lParam)) != NULL )
			{
				ed->app->CheckMenus() ;
				SetFocus(ed->app->wEditor) ;
				Active = ed ;
			}
			break ;

		case WM_SIZE:			// Must also be passed to DefMDIChildProc
			if (wParam != 1) 
			{
				RECT rc;
				GetClientRect(hWnd, &rc);
				if ( (ed = FindEditor(hWnd)) != NULL ) // ed = Active ;
					SetWindowPos(ed->app->wEditor, 0, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, 0);
			}

			break;

		case WM_COMMAND:
			if ( GetActiveEditor() != NULL ) Active->app->CheckMenus();
			break ;
//			return 0;

		case WM_NOTIFY:
			if ( (ed = FindEditor(hWnd)) != NULL )
				ed->app->Notify(reinterpret_cast<SCNotification *>(lParam));
			break ;

		case WM_MENUSELECT:
			if ( GetActiveEditor() != NULL ) 
			{
				Active->app->CheckMenus();
				SetFocus(Active->app->wEditor) ;
			}
			return 0;

		case WM_CHILDACTIVATE:	// Must also be passed to DefMDIChildProc
			if ( GetActiveEditor() != NULL ) 
			{
				Active->app->CheckMenus();
				SetFocus(Active->app->wEditor) ;
			}
			break ;

		case WM_CLOSE:
			if ( GetActiveEditor() != NULL )
			{
				if (Active->app->SaveIfUnsure() != IDCANCEL)
				{
					RemoveEditorWindow(Active->hWnd) ;
					GetActiveEditor() ;
				}
				if ( Active == NULL ) DisableMenu() ;
				return 0 ;
			}
			else
				DisableMenu() ;
 	}
	return DefMDIChildProc(hWnd, iMessage, wParam, lParam);
}


void CheckHelpCombos(int x)
{
	int cat, subcat ;

	if ( x == CBN_SELCHANGE )
	{
		cat = SendMessage(hwCat, CB_GETCURSEL, 0, 0); 
		subcat = SendMessage(hwSubCat, CB_GETCURSEL, 0, 0); 
		if ( cat != LastCat || subcat != LastSubCat )
		{
			SetKeywordsComboEx(cat,subcat) ;
			LastCat = cat ;
			LastSubCat = subcat ;
		}
	}
}


LRESULT CALLBACK HelpWndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{

	HWND chw = (HWND)lParam ;
	int x ;
	char kwname[MAX_PATH] ;
	const char *tip_usage, *tip_desc ;
	char *tmptd, *p ;


	switch (iMessage)
	{
		case WM_CREATE:
			DrawHelpWindow(hWnd) ;
			break ;

		case WM_MDIACTIVATE:
			break ;

		case WM_SIZE:			// Must also be passed to DefMDIChildProc
			if (wParam != 1) 
			{
				RECT rc;
				GetClientRect(hWnd, &rc);
				SetWindowPos(hwEdit, 0, 10, 110 , rc.right - rc.left - 20 , rc.bottom - rc.top - 120, 0);
			}
			break;

		case WM_COMMAND:
			x = wParam >> 16 ;
			if ( chw == hwCat || chw == hwSubCat )	// Category/sub-category combo box
			{
				CheckHelpCombos(x) ;
			}
			else if ( chw == hwBut )		// Show button
			{
				if ( x == BN_CLICKED )
				{
					SendMessage(hwKW, WM_GETTEXT, MAX_PATH, (LPARAM)kwname) ;
					if (ScriptXML.FindNode("keyword", "name", (const char *)kwname) == 0 )
					{
						ScriptXML.MoveToFirstChild() ;
						tip_usage = ScriptXML.NodeValue() ;
						ScriptXML.MoveToNextSibling() ;
						tip_desc = ScriptXML.NodeValue() ;
						tmptd = new char [strlen(tip_desc)+1] ;
						strcpy(tmptd, tip_desc) ;
						p = tmptd ;
						while (*p) 
						{
							if ( *p == '\\' && *(p+1) == 'n')
							{
								*p = 13 ;
								*(p+1) = 10 ;
							}
							p++ ;
						}
						if ( strcmp((const char *)RefCat[LastCat], "ARTICLE") == 0 )
							sprintf(RefDesc, "%s",tmptd) ;
						else
							sprintf(RefDesc, "Usage: %s\r\n\r\nDescription:\r\n%s", tip_usage, tmptd) ;
						SendMessage(hwEdit, WM_SETTEXT, 0, (LPARAM)RefDesc) ;
					}
				}
			}
			break ;

		case WM_NOTIFY:
			break ;

		case WM_MENUSELECT:
			return 0;

		case WM_CHILDACTIVATE:	// Must also be passed to DefMDIChildProc
			break ;

		case WM_CLOSE:
			SendMessage(hwMDIClient, WM_MDIDESTROY, (WPARAM)hwHelp, 0) ;
			hwHelp = NULL ;
 	}
	return DefMDIChildProc(hWnd, iMessage, wParam, lParam);
}



void SetKeywordsCombo(int cat, int subcat)
{
	int kwcount, i ;


	SendDlgItemMessage(hwRefDialogue, IDC_COMBO_KW, CB_RESETCONTENT, 0, 0); 
	ScriptXML.FindNode("keywords",(long)1) ;
	kwcount = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	for ( i = 0 ; i < kwcount ; i++ )
	{
		ScriptXML.FindNode("keyword",i+1) ;
		if (strcmp(RefCat[cat], ScriptXML.GetNodeAttrValue("category")) == 0 &&
			strcmp(RefSubCat[subcat], ScriptXML.GetNodeAttrValue("subcategory")) == 0 )
		{
			SendDlgItemMessage(hwRefDialogue, IDC_COMBO_KW, CB_ADDSTRING, 0, (LPARAM)ScriptXML.GetNodeAttrValue("name")); 
		}
	}
	SendDlgItemMessage(hwRefDialogue, IDC_COMBO_KW, CB_SETCURSEL, 0, 0); 
}


void SetKeywordsComboEx(int cat, int subcat)
{
	int kwcount, i ;


	SendMessage(hwKW, CB_RESETCONTENT, 0, 0); 
	ScriptXML.FindNode("keywords",(long)1) ;
	kwcount = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	for ( i = 0 ; i < kwcount ; i++ )
	{
		ScriptXML.FindNode("keyword",i+1) ;
		if (strcmp(RefCat[cat], ScriptXML.GetNodeAttrValue("category")) == 0 &&
			strcmp(RefSubCat[subcat], ScriptXML.GetNodeAttrValue("subcategory")) == 0 )
		{
			SendMessage(hwKW, CB_ADDSTRING, 0, (LPARAM)ScriptXML.GetNodeAttrValue("name")); 
		}
	}
	SendMessage(hwKW, CB_SETCURSEL, 0, 0); 
}


void InitialiseRef(void)
{
	int numcats, numsubcats, i ;

	if ( RefCat != NULL ) delete RefCat ;
	if ( RefSubCat != NULL ) delete RefSubCat ;

	hwRefDialogue = CreateDialog(hInst, MAKEINTRESOURCE(IDD_DLG_REFERENCE),hwMDIFrame, RefWndProc) ;

	SendDlgItemMessage(hwRefDialogue, IDC_COMBO_CAT, CB_RESETCONTENT, 0, 0); 
	ScriptXML.FindNode("categories",(long)1) ;
	numcats = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	RefCat = new const char * [numcats] ;
	for ( i = 0 ; i < numcats ; i++ )
	{
		ScriptXML.FindNode("category",i+1) ;
		RefCat[i] = ScriptXML.GetNodeAttrValue("name") ;
		SendDlgItemMessage(hwRefDialogue, IDC_COMBO_CAT, CB_ADDSTRING, 0, (LPARAM)ScriptXML.NodeValue()); 
	}

	SendDlgItemMessage(hwRefDialogue, IDC_COMBO_SUBCAT, CB_RESETCONTENT, 0, 0); 
	ScriptXML.FindNode("subcategories",(long)1) ;
	numsubcats = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	RefSubCat = new const char * [numsubcats] ;
	for ( i = 0 ; i < numsubcats ; i++ )
	{
		ScriptXML.FindNode("subcategory",i+1) ;
		RefSubCat[i] = ScriptXML.GetNodeAttrValue("name") ;
		SendDlgItemMessage(hwRefDialogue, IDC_COMBO_SUBCAT, CB_ADDSTRING, 0, (LPARAM)ScriptXML.NodeValue()); 
	}

	SendDlgItemMessage(hwRefDialogue, IDC_COMBO_CAT, CB_SETCURSEL, 0, 0); 
	SendDlgItemMessage(hwRefDialogue, IDC_COMBO_SUBCAT, CB_SETCURSEL, 0, 0); 
	SetKeywordsCombo(0,0) ;
	ShowWindow(hwRefDialogue, SW_SHOW) ;
}


/*
 * Load and parse script data from XML file. Sets up
 * keyword strings.
 */
bool LoadScriptXML(void)
{
	FILE *fd ;
	struct stat st ;
	char *buff ;
	int old_fmode ;
	CConcat factwords, actionwords, symbolwords, keywords, paramwords ;
	int kwcount ;
	long i ;
	const char *kwtype, *kwname ;
	char modpath[MAX_PATH+1] ;
	CRegistry reg ;


	if ( true == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY) )
	{
		if ( false == reg.GetValue("XMLFile", modpath, MAX_PATH) )
		{
			MessageBox(0, "Cannot find XML file registry setting","Error", MB_OK) ;
			return false ;
		}
	}


	old_fmode = _fmode ;
	_fmode = O_BINARY ;
	if ( stat((const char *)modpath, &st) != 0 )
	{
		MessageBox(0,"Failed to stat() Scripting XML data","Error", MB_OK) ;
		return false ;
	}

	if ( (fd = fopen((const char *)modpath, "r")) == NULL )
	{
		MessageBox(0,"Failed to fopen() Scripting XML data","Error", MB_OK) ;
		return false ;
	}

	buff = new char [st.st_size] ;
	if ( fread((void *)buff, st.st_size, 1, fd) != 1 )
	{
		MessageBox(0,"Failed to fread() Scripting XML data","Error", MB_OK) ;
		fclose(fd) ;
		delete buff ;
		return false ;
	}

	fclose(fd) ;
	
	ScriptXML.LoadData(st.st_size, buff) ;
	if ( ScriptXML.Parse() != 0 )
	{
		MessageBox(0,"Failed to parse Scripting XML data","Error", MB_OK) ;
		delete buff ;
		return false ;
	}

	delete buff ;
	_fmode = old_fmode ;

	ScriptXML.FindNode("keywords",(long)1) ;
	kwcount = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	for ( i = 0 ; i < kwcount ; i++ )
	{
		ScriptXML.FindNode("keyword",i+1) ;
		kwtype = ScriptXML.GetNodeAttrValue("category") ;
		kwname = ScriptXML.GetNodeAttrValue("name") ;

		if ( strcmp(kwtype, "FACT") == 0 )
		{
			factwords.Append(kwname) ;
			factwords.Append(" ") ;
		}
		else
		if ( strcmp(kwtype, "ACTION") == 0 )
		{
			actionwords.Append(kwname) ;
			actionwords.Append(" ") ;
		}
		else
		if ( strcmp(kwtype, "SYMBOL") == 0 )
		{
			symbolwords.Append(kwname) ;
			symbolwords.Append(" ") ;
		}
		else
		if ( strcmp(kwtype, "KEYWORD") == 0 )
		{
			keywords.Append(kwname) ;
			keywords.Append(" ") ;
		}
		else
		if ( strcmp(kwtype, "PARAM") == 0 )
		{
			paramwords.Append(kwname) ;
			paramwords.Append(" ") ;
		}

		if ( strcmp(kwtype, "ARTICLE") != 0 )
		{
			acwords.Append(kwname) ;
			acwords.Append(" ") ;
			AddACWord(kwname, true) ;
		}
	}

	AoCKeyWords = keywords.ToString() ;
	ActionKeyWords = actionwords.ToString() ;
	ParamKeyWords = paramwords.ToString() ;
	FactKeyWords = factwords.ToString() ;
	SysObjectWords = symbolwords.ToString() ;
	acWords = acwords.ToString() ;
	tmpacw = new char [strlen((const char *)acWords)+1] ;
	return true ;
}


bool AddACWord(const char *w, bool init)
{
	ACWORD *acwtmp ;
	int c ;


	if ( init == true )		// Words passed are already in lexical order
	{
		acwtmp = new ACWORD ;
		acwtmp->word = strdup(w) ;
		acwtmp->next = NULL ;
		acwtmp->prev = NULL ;
		if ( ACWTop == NULL )
			ACWTop = acwtmp ;
		else
		{
			ACWLast->next = acwtmp ;
			acwtmp->prev = ACWLast ;
		}
		ACWLast = acwtmp ;
		acwtmp->type = 'K' ;
	}
	else
	{
		acwtmp = ACWTop ;
		while ( (c = strcmp((const char *)acwtmp->word, w)) < 0)
		{
			if (acwtmp->next != NULL )
				acwtmp = acwtmp->next ;
			else
				break ;
		}

		// At this point we are either at the end of the list, 
		// have an exact match, or are just past our insertion point

		if ( c == 0 ) return false ; //already have the word

		if ( acwtmp->next == NULL ) // At end of list
		{
			acwtmp = new ACWORD ;
			acwtmp->type = 'U' ;		
			acwtmp->word = strdup(w) ;
			acwtmp->next = NULL ;
			ACWLast->next = acwtmp ;
			acwtmp->prev = ACWLast ;
			ACWLast = acwtmp ;		
		}
		else						// The current word is greater
		{
			ACWORD *newnode = new ACWORD ;
			newnode->word = strdup(w) ;
			newnode->type = 'U' ;		
			if ( acwtmp != ACWTop )
			{
				newnode->prev = acwtmp->prev ;
				acwtmp->prev->next = newnode ;
				acwtmp->prev = newnode ;
				newnode->next = acwtmp ;
			}
			else
			{
				ACWTop-> prev = newnode ;
				newnode->prev = NULL ;
				newnode->next = ACWTop ;
				ACWTop = newnode ;
			}
		}
	}
	return true ;		
}


void DestroyACWords(void)
{
	ACWORD *acwtmp ;

	for ( acwtmp = ACWLast ; acwtmp->prev != NULL ; )
	{
		acwtmp = acwtmp->prev ;
		delete acwtmp->next->word ;
		delete acwtmp->next ;
	}	
	delete ACWTop ;
	ACWTop = ACWLast = NULL ;
}


bool CheckXMLLocation(void)
{
	HMODULE hM ;
	char modpath[MAX_PATH+1], *p ;
	unsigned long l ;
	CRegistry reg ;
	struct stat st ;


	if ( true == reg.OpenKey(HKEY_CURRENT_USER, BASE_REG_KEY) )
	{
		if ( true == reg.GetValue("XMLFile", modpath, MAX_PATH) )
		{
			if ( stat((const char *)modpath, &st) == 0 ) return true ;
		}
	}

	if ( (hM = GetModuleHandle(NULL)) != NULL )
	{
		if ( (l = GetModuleFileName(hM, modpath, MAX_PATH)) != 0 )
		{
			p = modpath + l - 1 ;
			while ( p > modpath && *p != '\\' ) p-- ;
			if ( *p == '\\' ) p++ ;
			sprintf(p,"%s",scriptinfo) ;
			l = strlen((const char *)modpath) ;
			if ( stat((const char *)modpath, &st) == 0 )
			{
				reg.SetValue("XMLFile", (const char *)modpath, l) ;
				return true ;
			}
		}
	}

	if ( _getcwd(modpath, MAX_PATH) != NULL )
	{
		strcat(modpath, "\\") ;
		strcat(modpath, scriptinfo) ;
		l = strlen((const char *)modpath) ;
		if ( stat((const char *)modpath, &st) == 0 )
		{
			reg.SetValue("XMLFile", (const char *)modpath, l) ;
			return true ;
		}
	}

	if ( GetWindowsDirectory(modpath, MAX_PATH) != 0 )
	{
		strcat(modpath, "\\") ;
		strcat(modpath, scriptinfo) ;
		l = strlen((const char *)modpath) ;
		if ( stat((const char *)modpath, &st) == 0 )
		{
			reg.SetValue("XMLFile", (const char *)modpath, l) ;
			return true ;
		}
	}

	return false ;
}


void InitToolTips(HWND hWnd)
{
	hwToolTip = CreateWindowEx(	WS_EX_TOPMOST, TOOLTIPS_CLASS, NULL, WS_POPUP|TTS_NOPREFIX|TTS_ALWAYSTIP,
								CW_USEDEFAULT, CW_USEDEFAULT,CW_USEDEFAULT, CW_USEDEFAULT,
								hWnd,NULL,hInst,0) ;
}


void DisableMenu(void)
{
	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FILE_SAVE, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FILE_SAVE, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_UNDO, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_UNDO, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_REDO, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_REDO, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_CUT, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_CUT, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_PASTE, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_PASTE, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_COPY, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_COPY, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_CLEAR_DEFCONSTS, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_CLEAR_DEFCONSTS, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_RESCAN_DEFCONSTS, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_RESCAN_DEFCONSTS, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FILE_PROPERTIES, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FILE_SAVEAS, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_CLOSE, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_DELETE, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_SELECTALL, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_GOTO_LINE, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_GOTO_LINE, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FIND, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FIND, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FIND_NEXT, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FIND_NEXT, FALSE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_REPLACE, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_REPLACE, FALSE) ;


}


void EnableMenu(void)
{
	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FILE_SAVE, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FILE_SAVE, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_CUT, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_CUT, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_PASTE, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_PASTE, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_COPY, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_EDIT_COPY, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_CLEAR_DEFCONSTS, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_CLEAR_DEFCONSTS, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_RESCAN_DEFCONSTS, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_RESCAN_DEFCONSTS, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FILE_PROPERTIES, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FILE_PROPERTIES, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FILE_SAVEAS, MF_ENABLED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_CLOSE, MF_ENABLED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_DELETE, MF_ENABLED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_EDIT_SELECTALL, MF_ENABLED | MF_BYCOMMAND);

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_GOTO_LINE, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_GOTO_LINE, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FIND, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FIND, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_FIND_NEXT, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_FIND_NEXT, TRUE) ;

	EnableMenuItem(GetMenu(hwMDIFrame), IDM_REPLACE, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(hwToolBar, TB_ENABLEBUTTON, IDM_REPLACE, TRUE) ;

}


void AddToolTipString(int id, const char *s)
{
	TOOLINFO ti ;
	RECT rect ;
	LRESULT l ;

	l = SendMessage(hwToolBar, TB_GETITEMRECT, id, (LPARAM)&rect);
	ti.cbSize = sizeof TOOLINFO ;
	ti.uFlags = TTF_CENTERTIP|TTF_SUBCLASS ;
	ti.hinst = hInst ;
	ti.hwnd = hwToolBar ;
	ti.lpszText = strdup(s) ;
	ti.rect.bottom = rect.bottom ;
	ti.rect.left = rect.left ;
	ti.rect.right = rect.right ;
	ti.rect.top = rect.top ;
	l = SendMessage(hwToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
	delete ti.lpszText ;
}


void InitStatusBar(HWND hWnd)
{
	int statwidths[] = {75,150,250,-1} ;
	WPARAM parts = sizeof(statwidths)/sizeof(int) ;

	hwStatusBar = CreateWindowEx(0, STATUSCLASSNAME, NULL, WS_CHILD|WS_VISIBLE|SBARS_SIZEGRIP,0,0,0,0,hWnd,NULL,hInst,0) ;
	SendMessage(hwStatusBar, SB_SETPARTS, parts, (LPARAM)statwidths);
}


void InitToolBar(HWND hWnd)
{
	TBBUTTON tbb[NUM_TB_BUTTONS] ;
	TBADDBITMAP tbab ;

	hwToolBar = CreateWindowEx(0,TOOLBARCLASSNAME, NULL, WS_CHILD|WS_VISIBLE,0,0,0,0,hWnd,NULL,hInst,0) ;
	SendMessage(hwToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0) ;

	tbab.hInst = hInst ;
	tbab.nID = IDB_BITMAP_NEW ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 0
	tbab.nID = IDB_BITMAP_OPEN ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 1
	tbab.nID = IDB_BITMAP_SAVE ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 2
	tbab.nID = IDB_BITMAP_UNDO ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 3
	tbab.nID = IDB_BITMAP_REDO ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 4
	tbab.nID = IDB_BITMAP_CUT ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 5
	tbab.nID = IDB_BITMAP_COPY ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 6
	tbab.nID = IDB_BITMAP_PASTE ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 7
	tbab.nID = IDB_BITMAP_LEFT ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 8
	tbab.nID = IDB_BITMAP_RIGHT ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 9
	tbab.nID = IDB_BITMAP_HTILE ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 10
	tbab.nID = IDB_BITMAP_VTILE ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 11
	tbab.nID = IDB_BITMAP_CASC ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 12
	tbab.nID = IDB_BITMAP_FINDTEXT ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 13
	tbab.nID = IDB_BITMAP_FIND_NEXT ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 14
	tbab.nID = IDB_BITMAP_REPLACE ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 15
	tbab.nID = IDB_BITMAP_GOTO ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 16
	tbab.nID = IDB_BITMAP_REMDC ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 17
	tbab.nID = IDB_BITMAP_SCANDC ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 18
	tbab.nID = IDB_BITMAP_SETTINGS ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 19
	tbab.nID = IDB_BITMAP_HELP ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 20
	tbab.nID = IDB_BITMAP_REF ;
	SendMessage(hwToolBar, TB_ADDBITMAP, 1, (LPARAM)&tbab);		// 21

	ZeroMemory(tbb, sizeof(tbb));

	tbb[0].iBitmap = 0 ;
    tbb[0].fsState = TBSTATE_ENABLED ;
    tbb[0].fsStyle = TBSTYLE_BUTTON ;
    tbb[0].idCommand = IDM_FILE_NEW ;

	tbb[1].iBitmap = 1 ;
    tbb[1].fsState = TBSTATE_ENABLED ;
    tbb[1].fsStyle = TBSTYLE_BUTTON ;
    tbb[1].idCommand = IDM_FILE_OPEN ;

	tbb[2].iBitmap = 2 ;
    tbb[2].fsState = TBSTATE_ENABLED ;
    tbb[2].fsStyle = TBSTYLE_BUTTON ;
    tbb[2].idCommand = IDM_FILE_SAVE ;

	tbb[3].iBitmap = -1 ;
    tbb[3].fsStyle = TBSTYLE_SEP ;

	tbb[4].iBitmap = 3 ;
    tbb[4].fsState = TBSTATE_ENABLED ;
    tbb[4].fsStyle = TBSTYLE_BUTTON ;
    tbb[4].idCommand = IDM_EDIT_UNDO ;

	tbb[5].iBitmap = 4 ;
    tbb[5].fsState = TBSTATE_ENABLED ;
    tbb[5].fsStyle = TBSTYLE_BUTTON ;
    tbb[5].idCommand = IDM_EDIT_REDO ;

	tbb[6].iBitmap = -1 ;
    tbb[6].fsStyle = TBSTYLE_SEP ;

	tbb[7].iBitmap = 5 ;
    tbb[7].fsState = TBSTATE_ENABLED ;
    tbb[7].fsStyle = TBSTYLE_BUTTON ;
    tbb[7].idCommand = IDM_EDIT_CUT ;

	tbb[8].iBitmap = 6 ;
    tbb[8].fsState = TBSTATE_ENABLED ;
    tbb[8].fsStyle = TBSTYLE_BUTTON ;
    tbb[8].idCommand = IDM_EDIT_COPY ;

	tbb[9].iBitmap = 7 ;
    tbb[9].fsState = TBSTATE_ENABLED ;
    tbb[9].fsStyle = TBSTYLE_BUTTON ;
    tbb[9].idCommand = IDM_EDIT_PASTE ;

	tbb[10].iBitmap = -1 ;
    tbb[10].fsStyle = TBSTYLE_SEP ;

	tbb[11].iBitmap = 8 ;
    tbb[11].fsState = TBSTATE_ENABLED ;
    tbb[11].fsStyle = TBSTYLE_BUTTON ;
    tbb[11].idCommand = IDM_MDI_PREV ;

	tbb[12].iBitmap = 9 ;
    tbb[12].fsState = TBSTATE_ENABLED ;
    tbb[12].fsStyle = TBSTYLE_BUTTON ;
    tbb[12].idCommand = IDM_MDI_NEXT ;

	tbb[13].iBitmap = 10 ;
    tbb[13].fsState = TBSTATE_ENABLED ;
    tbb[13].fsStyle = TBSTYLE_BUTTON ;
    tbb[13].idCommand = IDM_WIN_TILE_H ;

	tbb[14].iBitmap = 11 ;
    tbb[14].fsState = TBSTATE_ENABLED ;
    tbb[14].fsStyle = TBSTYLE_BUTTON ;
    tbb[14].idCommand = IDM_WIN_TILE_V ;

	tbb[15].iBitmap = 12 ;
    tbb[15].fsState = TBSTATE_ENABLED ;
    tbb[15].fsStyle = TBSTYLE_BUTTON ;
    tbb[15].idCommand = IDM_WIN_CASCADE ;

	tbb[16].iBitmap = -1 ;
    tbb[16].fsStyle = TBSTYLE_SEP ;

	tbb[17].iBitmap = 13 ;
    tbb[17].fsState = TBSTATE_ENABLED ;
    tbb[17].fsStyle = TBSTYLE_BUTTON ;
    tbb[17].idCommand = IDM_FIND ;

	tbb[18].iBitmap = 14 ;
    tbb[18].fsState = TBSTATE_ENABLED ;
    tbb[18].fsStyle = TBSTYLE_BUTTON ;
    tbb[18].idCommand = IDM_FIND_NEXT ;

	tbb[19].iBitmap = 15 ;
    tbb[19].fsState = TBSTATE_ENABLED ;
    tbb[19].fsStyle = TBSTYLE_BUTTON ;
    tbb[19].idCommand = IDM_REPLACE ;

	tbb[20].iBitmap = -1 ;
    tbb[20].fsStyle = TBSTYLE_SEP ;

	tbb[21].iBitmap = 16 ;
    tbb[21].fsState = TBSTATE_ENABLED ;
    tbb[21].fsStyle = TBSTYLE_BUTTON ;
    tbb[21].idCommand = IDM_GOTO_LINE ;

	tbb[22].iBitmap = -1 ;
    tbb[22].fsStyle = TBSTYLE_SEP ;

	tbb[23].iBitmap = 17 ;
    tbb[23].fsState = TBSTATE_ENABLED ;
    tbb[23].fsStyle = TBSTYLE_BUTTON ;
    tbb[23].idCommand = IDM_CLEAR_DEFCONSTS ;

	tbb[24].iBitmap = 18 ;
    tbb[24].fsState = TBSTATE_ENABLED ;
    tbb[24].fsStyle = TBSTYLE_BUTTON ;
    tbb[24].idCommand = IDM_RESCAN_DEFCONSTS ;

	tbb[25].iBitmap = -1 ;
    tbb[25].fsStyle = TBSTYLE_SEP ;

	tbb[26].iBitmap = 19 ;
    tbb[26].fsState = TBSTATE_ENABLED ;
    tbb[26].fsStyle = TBSTYLE_BUTTON ;
    tbb[26].idCommand = IDM_SETTINGS ;

	tbb[27].iBitmap = 20 ;
    tbb[27].fsState = TBSTATE_ENABLED ;
    tbb[27].fsStyle = TBSTYLE_BUTTON ;
    tbb[27].idCommand = IDM_TIPS ;

	tbb[28].iBitmap = -1 ;
    tbb[28].fsStyle = TBSTYLE_SEP ;

	tbb[29].iBitmap = 21 ;
    tbb[29].fsState = TBSTATE_ENABLED ;
    tbb[29].fsStyle = TBSTYLE_BUTTON ;
    tbb[29].idCommand = IDM_HELP_REF ;

	SendMessage(hwToolBar, TB_ADDBUTTONS, sizeof(tbb)/sizeof(TBBUTTON), (LPARAM)&tbb);
	SendMessage(hwToolBar, TB_AUTOSIZE, 0, 0);
	ShowWindow(hwToolBar, SW_SHOW) ;

	AddToolTipString(0, "New") ;
	AddToolTipString(1, "Open") ;
	AddToolTipString(2, "Save") ;
	AddToolTipString(4, "Undo") ;
	AddToolTipString(5, "Redo") ;
	AddToolTipString(7, "Cut") ;
	AddToolTipString(8, "Copy") ;
	AddToolTipString(9, "Paste") ;
	AddToolTipString(11, "Previous window") ;
	AddToolTipString(12, "Next window") ;
	AddToolTipString(13, "Tile horizontally") ;
	AddToolTipString(14, "Tile vertically") ;
	AddToolTipString(15, "Cascade") ;
	AddToolTipString(17, "Find") ;
	AddToolTipString(18, "Find next") ;
	AddToolTipString(19, "Replace") ;
	AddToolTipString(21, "Goto line") ;
	AddToolTipString(23, "Clear defconsts") ;
	AddToolTipString(24, "Rescan defconsts") ;
	AddToolTipString(26, "Settings") ;
	AddToolTipString(27, "Tips") ;
	AddToolTipString(29, "Scripting reference") ;
	SendMessage(hwToolTip, TTM_ACTIVATE, TRUE, 0) ;
	ShowWindow(hwToolTip, SW_SHOW) ;
}


/*
 * Register the MDI frame window class
 */
void RegisterMDIWindowClass(HINSTANCE hInstance) 
{
	const char resourceName[] = "DMApp";
	WNDCLASSEX wndclass;


	wndclass.cbSize = sizeof WNDCLASSEX ;
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = MDIFrameWndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1)) ;
	wndclass.hCursor = NULL;
	wndclass.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
	wndclass.lpszMenuName = resourceName;
	wndclass.lpszClassName = MDIclassName;
	wndclass.hIconSm = NULL ;

	if (!::RegisterClassEx(&wndclass))
		::exit(FALSE);
}


/*
 * Register the MDI child window class
 */
void RegisterWindowClass(HINSTANCE hInstance) 
{
	WNDCLASSEX wndclass;


	wndclass.cbSize = sizeof WNDCLASSEX ;
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON2)) ;
	wndclass.hCursor = NULL;
	wndclass.hbrBackground = NULL;
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = className;
	wndclass.hIconSm = NULL ;

	if (!::RegisterClassEx(&wndclass))
		::exit(FALSE);
}


void RegisterHelpWindowClass(HINSTANCE hInstance) 
{
	WNDCLASSEX wndclass;


	wndclass.cbSize = sizeof WNDCLASSEX ;
	wndclass.style = CS_HREDRAW | CS_VREDRAW ;
	wndclass.lpfnWndProc = HelpWndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON3)) ;
	wndclass.hCursor = NULL;
	wndclass.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = HelpClassName;
	wndclass.hIconSm = NULL ;

	if (!::RegisterClassEx(&wndclass))
		::exit(FALSE);
}

bool CreateHelpWindow(void)
{
	hwHelp = CreateWindowEx(
	            WS_EX_CLIENTEDGE|WS_EX_MDICHILD,
		        HelpClassName,
			    "Script Reference",
				WS_CAPTION | WS_SYSMENU | WS_THICKFRAME |
			    WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_CLIPCHILDREN,
				CW_USEDEFAULT, CW_USEDEFAULT, 
				580, 550, 
				hwMDIClient,
				NULL,
				hInst,
				0);
	return hwHelp == NULL ? false : true ;
}


bool DrawHelpWindow(HWND hWnd)
{
	int numcats, numsubcats, i ;

	if ( RefCat != NULL ) delete RefCat ;
	if ( RefSubCat != NULL ) delete RefSubCat ;

	CreateWindow("STATIC", "Category",WS_VISIBLE|WS_CHILD|SS_RIGHT,10,10,100,30,hWnd,NULL,hInst,NULL);
	CreateWindow("STATIC", "Sub-category",WS_VISIBLE|WS_CHILD|SS_RIGHT,10,40,100,30,hWnd,NULL,hInst,NULL);
	CreateWindow("STATIC", "Keyword",WS_VISIBLE|WS_CHILD|SS_RIGHT,10,70,100,30,hWnd,NULL,hInst,NULL);

	hwCat = CreateWindow("COMBOBOX", "",WS_VISIBLE|WS_VSCROLL|WS_CHILD|CBS_DROPDOWN ,120,10,150,300,hWnd,NULL,hInst,NULL);

	SendMessage(hwCat,CB_RESETCONTENT, 0, 0); 
	ScriptXML.FindNode("categories",(long)1) ;
	numcats = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	RefCat = new const char * [numcats] ;
	for ( i = 0 ; i < numcats ; i++ )
	{
		ScriptXML.FindNode("category",i+1) ;
		RefCat[i] = ScriptXML.GetNodeAttrValue("name") ;
		SendMessage(hwCat, CB_ADDSTRING, 0, (LPARAM)ScriptXML.NodeValue()); 
	}

	hwSubCat = CreateWindow("COMBOBOX", "",WS_VISIBLE|WS_VSCROLL|WS_CHILD|CBS_DROPDOWN ,120,40,200,350,hWnd,NULL,hInst,NULL);
	SendMessage(hwSubCat, CB_RESETCONTENT, 0, 0); 
	ScriptXML.FindNode("subcategories",(long)1) ;
	numsubcats = atoi(ScriptXML.GetNodeAttrValue("count")) ;
	RefSubCat = new const char * [numsubcats] ;
	for ( i = 0 ; i < numsubcats ; i++ )
	{
		ScriptXML.FindNode("subcategory",i+1) ;
		RefSubCat[i] = ScriptXML.GetNodeAttrValue("name") ;
		SendMessage(hwSubCat, CB_ADDSTRING, 0, (LPARAM)ScriptXML.NodeValue()); 
	}

	SendMessage(hwCat, CB_SETCURSEL, 0, 0); 
	SendMessage(hwSubCat, CB_SETCURSEL, 0, 0); 

	hwKW = CreateWindow("COMBOBOX", "",WS_VISIBLE|WS_VSCROLL|WS_CHILD|CBS_DROPDOWN ,120,70,300,350,hWnd,NULL,hInst,NULL);
	SetKeywordsComboEx(0,0) ;

	hwEdit = CreateWindowEx(WS_EX_CLIENTEDGE,"EDIT", "", WS_VISIBLE|WS_VSCROLL|WS_HSCROLL|WS_CHILD|WS_BORDER|ES_MULTILINE, 10, 110, 550, 400, hWnd,NULL,hInst,NULL) ;

	hwBut = CreateWindow("BUTTON", "Show", WS_VISIBLE|WS_CHILD|BS_DEFPUSHBUTTON,430, 70, 50, 25, hWnd,NULL,hInst,NULL) ;

	return true ;
}


/*
 * Main entry point for kicking everything off
 */
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE, LPSTR lpszCmdLine, int nCmdShow) 
{

	hInst = hInstance ;
	HACCEL hAccTable = LoadAccelerators(hInstance, "DMApp");
	DWORD lasterr ;
	int i ;


	UserKeyWords = NULL ;
	TipsSwitchedOn = true ;
	AllowAutoC = BST_CHECKED ;
	RefCat = RefSubCat = NULL ;
	for ( i = 0 ; i < MAX_RECENT ; i++ )
	{
		rfptr[i] = NULL ;
		Recent[i].ed = NULL ;
		Recent[i].id = 0 ;
		Recent[i].DisplayName[0] = 0 ;
		Recent[i].FullPath[0] = 0 ;
	}

	pbinf[0].id = IDC_CC_PB1 ;
	pbinf[1].id = IDC_CC_PB2 ;
	pbinf[2].id = IDC_CC_PB3 ;
	pbinf[3].id = IDC_CC_PB4 ;
	pbinf[4].id = IDC_CC_PB5 ;
	pbinf[5].id = IDC_CC_PB6 ;
	pbinf[6].id = IDC_CC_PB7 ;
	pbinf[7].id = IDC_CC_PB8 ;
	pbinf[8].id = IDC_CC_PB9 ;
	pbinf[9].id = IDC_CC_PB10 ;
	pbinf[10].id = IDC_CC_PB11 ;

	if ( LoadLibrary("SciLexer.DLL") == NULL )
	{
		MessageBox(0,"Cannot load SciLexer.DLL","Fatal error", MB_OK) ;
		return 0 ;
	}

	RegisterMDIWindowClass(hInstance);
	RegisterWindowClass(hInstance);
	RegisterHelpWindowClass(hInstance);
    GetRegistrySettings() ;
	fnbuff[0] = fText[0] = rText[0] = 0 ;
    if ( lpszCmdLine != NULL )
	{
		if (strlen((const char *)lpszCmdLine) > 0 )
		{
			strcpy(fnbuff, (const char *)lpszCmdLine) ;
		}
	}
	if ( CheckXMLLocation() == false )
	{
		MessageBox(0,"Cannot find scriptinfo XML data file","File Error", MB_OK) ;
		return 1 ;
	}

	if ( LoadScriptXML() == false ) return 1 ;

	InitCommonControls() ;		// Basically a LoadLibrary() for COMCTL32.DLL

	// Create MDI frame window
	hwMDIFrame = CreateWindowEx(
	             WS_EX_CLIENTEDGE,
	             MDIclassName,
	             "ScriptEd - AI Scripters script editor",
	             WS_CAPTION | WS_SYSMENU | WS_THICKFRAME |
	             WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
	             WS_MAXIMIZE | WS_CLIPCHILDREN,
				 CW_USEDEFAULT, CW_USEDEFAULT, 
				 CW_USEDEFAULT, CW_USEDEFAULT, 
	             0,
	             0, 
	             hInstance,
	             0);

	if ( hwMDIFrame == NULL )
	{
		lasterr = GetLastError() ;
		MessageBox(0,"Failed to create MDI frame window","ScriptEd",MB_OK) ;
		return 1 ;
	}

	ShowWindow(hwMDIFrame, SW_SHOW) ;

	do { Sleep(250) ; } while ( gotmdi == false ) ;
//	InitialiseRef() ;

	GetRegRecentFiles() ;

	if ( fnbuff[0] == 0 )
	{
		if ( rfcount > 0 )
			strcpy(fnbuff, (const char *)Recent[0].FullPath) ;
	}


	if ( fnbuff[0] != 0 ) 
	{
		NewEditorWindow() ;
		ScanFileDefconsts(Active) ;
	}
	else
		DisableMenu() ;

	if ( TipsSwitchedOn == true )
		SendMessage(hwStatusBar, SB_SETTEXT, 2, (LPARAM)"Tips ON") ;
	else
		SendMessage(hwStatusBar, SB_SETTEXT, 2, (LPARAM)"Tips OFF") ;

	// Message translation and dispatch loop.
	bool going = true;
	MSG msg;
	msg.wParam = 0;
	while (going) 
	{
		going = (GetMessage(&msg, NULL, 0, 0) == 0 ? false : true);
		if ( going )
		{
			if ( hwCurrentDialogue != NULL)
			{
				if (!IsDialogMessage(hwCurrentDialogue, &msg)) 
				{
					if (!TranslateMDISysAccel(hwMDIClient, &msg) && 
						!TranslateAccelerator(hwMDIFrame, hAccTable, &msg))
					{ 
						TranslateMessage(&msg); 
						DispatchMessage(&msg); 
					} 
				}
			}
			else
			{
				if (!IsDialogMessage(hwRefDialogue, &msg)) 
				{
					if (!TranslateMDISysAccel(hwMDIClient, &msg) && 
						!TranslateAccelerator(hwMDIFrame, hAccTable, &msg))
					{ 
						TranslateMessage(&msg); 
						DispatchMessage(&msg); 
					} 
				}
			}
		}
	}

	SetRegRecentFiles() ;
	// Tidy up
	if (RefCat != NULL) delete RefCat ;
	if (RefCat != NULL) delete RefSubCat ;
	DestroyACWords() ;
	delete AoCKeyWords ;
	delete ActionKeyWords ;
	delete ParamKeyWords ;
	delete FactKeyWords ;
	delete SysObjectWords ;
	delete acWords ;
	delete tmpacw ;
	delete UserKeyWords ;
	return msg.wParam;
}

/*

#include "SampleApp.h"

CColourPicker *cpicker;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
	{
	InitCommonControls();

	DialogBox(hInstance, MAKEINTRESOURCE(IDD_DIALOG), NULL, (DLGPROC)DlgProc);
    return 1;	
	}

LRESULT CALLBACK DlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
	switch(uMsg)
		{
		case WM_INITDIALOG:
			{
			try
				{
				unsigned short r=150, g=190, b=240, a=100;

				cpicker = new CColourPicker(hDlg, r, g, b, a, true);
				}
			catch(...)
				{
				MessageBox(hDlg, "Cannot create colour picker!", "Fatal error.", MB_OK | MB_ICONERROR);
				ExitProcess(-1);
				}
			}
		break;

		case WM_CTLCOLORSTATIC:
			{
			SColour colour = cpicker->GetCurrentColour();

			DrawCheckedRect(GetDlgItem(hDlg, IDC_COLOUR), colour.r, colour.g, colour.b, colour.a, 10, 10);
			}
		break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
				{
				case IDB_QUIT:
					{
					EndDialog(hDlg, IDB_QUIT);
					}
				break;

				case IDB_TEST:
					{
					short flag;
					flag = (SendDlgItemMessage(hDlg, IDC_ALPHA, BM_GETCHECK, 0, 0) == BST_CHECKED) ? 
						CP_USE_ALPHA : CP_NO_ALPHA;
					
					cpicker->CreatecolourPicker(flag);
					SendDlgItemMessage(hDlg, IDC_COLOUR, WM_PAINT, 0, 0);
					}
 				break;
				}
		break;
		}
	return 0L;
	}

*/