/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/

#ifndef _SCREDIT_H_
#define _SCREDIT_H_

#include <windows.h>		// Required for definition of HWND
#include "cpicker.h"		// nICO's colour picker control

// Length of buffer for storing font names
#define FONT_NAME_SIZE		64

// Base registry key where we store our settings
#define BASE_REG_KEY		"Software\\AIScripters\\ScriptEd"

// Position on menu of the File sub-menu
#define FILEMENU			0
#define IDM_FILESTART		45000
#define MAX_RECENT			10
#define MAX_DISPLAY_SIZE	40

// Position on menu of the Tools sub-menu
#define TOOLSMENU			2

// Position on menu of the Window sub-menu
#define WINDOWMENU			3

// MDI child window identifiers start from this number
#define IDM_WINDOWCHILD		20000

// Reference item description buffer size
#define REF_DESC_SIZE		32768

// size of work buffer 
#define WORK_BUFF_SIZE		4096

// Number of toolbar buttons
#define NUM_TB_BUTTONS		30

// Max length of a status bar item
#define SB_ITEM_LENGTH		127

// Number of customisable colours
#define MAX_COLOURS			11


// Recent file menu identifiers
#define IDM_RF1				(IDM_FILESTART+1)
#define IDM_RF2				(IDM_FILESTART+2)
#define IDM_RF3				(IDM_FILESTART+3)
#define IDM_RF4				(IDM_FILESTART+4)
#define IDM_RF5				(IDM_FILESTART+5)
#define IDM_RF6				(IDM_FILESTART+6)
#define IDM_RF7				(IDM_FILESTART+7)
#define IDM_RF8				(IDM_FILESTART+8)
#define IDM_RF9				(IDM_FILESTART+9)
#define IDM_RF10			(IDM_FILESTART+10)


const char MDIclassName[]	= "MDIScrEAppWindow";
const char className[]		= "ScrEAppWindow";
const char scriptinfo[]		= "scriptinfo.xml" ;
const char HelpClassName[]	= "HelpAppWindow" ;
const char credits[]		= "Credits to:\r\n\r\nEnsemble Studios\r\nMicrosoft Corp.\r\nwww.aiscripters.com\r\n\tRobink\r\n\tVNS_Caesar\r\n\tElite Raider\r\n\tKova\r\n\tVader\r\n\tMaximus\r\n\tMabuse\r\n\tTea315\r\nwww.scintilla.org\r\n\tNeil Hodgson\r\n\tPaul Winwood\r\nwww.mrfixitonline.com\r\nsourceforge.com\r\ncodeproject.com\r\n\tnICO" ;

// structure to hold a linked list of editor window parameters
typedef struct EditorWindow
{
	HWND hWnd ;
	ScrEApp *app ;
	void *pDoc ;
	struct EditorWindow *next ;
	struct EditorWindow *prev ;
} EDWIN ;


// structure to hold defconst words
typedef struct AutoCompleteWords
{
	char *word ;
	struct AutoCompleteWords *next ;
	struct AutoCompleteWords *prev ;
	char type ;
} ACWORD ;


// structure to store colour names and RGB values
typedef struct SEColours
{
	char *name ;
	COLORREF colour ;
} SECOLOUR ;


typedef struct pbInfo
{
	HWND hWnd ;
	int id ;
	COLORREF cref ;
} PBINFO ;

// Structure to store recent file info
typedef struct RecentFile
{
	char FullPath[MAX_PATH] ;
	char DisplayName[MAX_DISPLAY_SIZE+1] ;
	EDWIN *ed ;
	unsigned int id ;
} RECENTFILE ;


// Function prototypes
void SetRegistrySettings(void) ;
void GetRegistrySettings(void) ;
EDWIN *NewEditorWindow(void) ;
EDWIN *FindEditor(HWND) ;
void RemoveEditorWindow(HWND) ;
EDWIN *GetActiveEditor(void) ;
int FindSearchText(int) ;
void InitGotoDlgOptions(HWND) ;
void InitFindDlgOptions(HWND) ;
void InitReplaceDlgOptions(HWND) ;
int __stdcall ScreenFontEnumProc(const LOGFONT *, const TEXTMETRIC *, DWORD, LPARAM) ;
void InitSettingsDlgOptions(HWND) ;
void GetSettingsDlgOptions(HWND) ;
void GetGotoDlgOptions(HWND) ;
void GetFindDlgOptions(HWND) ;
void GetReplaceDlgOptions(HWND) ;
int __stdcall GotoWndProc(HWND, UINT, WPARAM, LPARAM) ;
int __stdcall FindWndProc(HWND, UINT, WPARAM, LPARAM) ;
void ApplySettings(void) ;
bool CloseAllWindows(void) ;
void RestoreChildren(void) ;
int __stdcall SettingsWndProc(HWND, UINT, WPARAM, LPARAM) ;
int __stdcall ReplaceWndProc(HWND, UINT, WPARAM, LPARAM) ;
LRESULT CALLBACK MDIFrameWndProc(HWND, UINT, WPARAM, LPARAM) ;
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM) ;
LRESULT CALLBACK HelpWndProc(HWND, UINT, WPARAM, LPARAM) ;
bool LoadScriptXML(void) ;
void RegisterMDIWindowClass(HINSTANCE) ;
void RegisterWindowClass(HINSTANCE) ;
int __stdcall AboutWndProc(HWND, UINT, WPARAM, LPARAM) ;
int __stdcall RefWndProc(HWND, UINT, WPARAM, LPARAM) ;
void InitialiseRef(void) ;
void SetKeywordsCombo(int, int) ;
void UpdateFileMenu(const char *, EDWIN *) ;
EDWIN *AlreadyOpen(const char *) ;
void SetRegRecentFiles(void) ;
void GetRegRecentFiles(void) ;
void OpenRecentFile(unsigned int) ;
bool CheckXMLLocation(void) ;
void GetFileStats(void) ;
void ScanFileDefconsts(EDWIN *) ;
bool AddACWord(const char *, bool) ;
void DestroyACWords(void) ;
void UpdateACWords(void) ;
void RescanDefconsts(void) ;
void RemoveDefconsts(void) ;
void InitToolBar(HWND) ;
void InitStatusBar(HWND) ;
void InitToolTips(HWND) ;
void AddToolTipString(int, const char *) ;
void DisableMenu(void) ;
void EnableMenu(void) ;
void RegisterHelpWindowClass(HINSTANCE) ;
bool CreateHelpWindow(void) ;
bool DrawHelpWindow(HWND) ;
void SetKeywordsComboEx(int, int) ;
void CheckHelpCombos(int) ;
void DoColourPicker(HWND, int) ;
int GetPBControlID(HWND) ;
void SetDefaultPBColour(int, SColour *) ;

#endif
