/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/

#ifndef _XMLPARSE_H_
#define _XMLPARSE_H_


#define  XML_VERSION_STRING    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"

/* XML parsed document node types */

#define NODE_BRANCH      1
#define NODE_LEAF        2
#define NODE_ATTR        3
#define NODE_PI          4  /* sequence <? .... ?>   */
#define NODE_DOCTYPE     5
#define NODE_FRAG        6
#define NODE_ATTRVAL     7
#define NODE_COMMENT     8  /* sequence <!-- .... -->   */
#define NODE_CDATA       9
#define NODE_EMPTY      10
#define NODE_UNDEFINED  -1


/* DTD node identifiers */

#define DTD_TOP            0
#define NODE_ENTITY        1
#define NODE_ELEMENT       2
#define NODE_ATTLIST       3
#define NODE_CONTENT_TYPE  4
#define NODE_CONTENT_MODEL 5
#define NODE_ATTRIBUTE_DEF 6
#define NODE_ATTRIBUTE_VAL 7

/* DTD Element, Attribute and other identifiers */

#define ELEMENT_ONLY_ONCE         0   /* No element qualifier */
#define ELEMENT_ONE_OR_MORE       1   /* '+' element qualifier */
#define ELEMENT_ZERO_OR_MORE      2   /* '*' element qualifier */
#define ELEMENT_ZERO_OR_ONE       3   /* '?' element qualifier */

#define ELEMENT_CONTENT_UNKNOWN  -1
#define ELEMENT_CONTENT_EMPTY     0   /* Element has no content eg. <br> */
#define ELEMENT_CONTENT_ANY       1   /* Element can contain anything eg. <html> */
#define ELEMENT_CONTENT_MIXED     2   /* Element can contain mixed data eg. <p> */
#define ELEMENT_CONTENT_KIDS      3   /* Element only has children eg. <table> */
#define ELEMENT_CONTENT_PCDATA    4

#define CONTENT_IS_PCDATA         (struct _BaseDTD *)0xFFFFFFFF

#define CONTENT_MODEL_CHOICE      1
#define CONTENT_MODEL_SEQ         2
#define CONTENT_MODEL_SINGLE      3   /* Single element so neither CHOICE  nor SEQ : this is a Berryism */

/* Defines for handling Attributes */

#define ATTRIBUTE_TYPE_STRING     1
#define ATTRIBUTE_TYPE_TOKEN      2
#define ATTRIBUTE_TYPE_ENUM       3

#define TOKEN_ID                  1
#define TOKEN_IDREF               2
#define TOKEN_IDREFS              3
#define TOKEN_ENTITY              4
#define TOKEN_ENTITIES            5
#define TOKEN_NMTOKEN             6
#define TOKEN_NMTOKENS            7

#define ATTR_STRING               "CDATA "
#define LEN_ATTR_STRING           6
#define TOKEN_ID_STR              "ID "
#define LEN_TOKEN_ID_STR          3
#define TOKEN_IDREF_STR           "IDREF "
#define LEN_TOKEN_IDREF_STR       6
#define TOKEN_IDREFS_STR          "IDREFS "
#define LEN_TOKEN_IDREFS_STR      7
#define TOKEN_ENTITY_STR          "ENTITY "
#define LEN_TOKEN_ENTITY_STR      7
#define TOKEN_ENTITIES_STR        "ENTITIES "
#define LEN_TOKEN_ENTITIES_STR    9
#define TOKEN_NMTOKEN_STR         "NMTOKEN "
#define LEN_TOKEN_NMTOKEN_STR     8
#define TOKEN_NMTOKENS_STR        "NMTOKENS "
#define LEN_TOKEN_NMTOKENS_STR    9

#define ATTR_REQUIRED             1
#define ATTR_IMPLIED              2
#define ATTR_FIXED                3
#define ATTR_ENUM                 4
#define ATTR_VALUE                5

#define ATTR_REQUIRED_STR         "#REQUIRED"
#define LEN_ATTR_REQUIRED_STR     9
#define ATTR_IMPLIED_STR          "#IMPLIED"
#define LEN_ATTR_IMPLIED_STR      8
#define ATTR_FIXED_STR            "#FIXED"
#define LEN_ATTR_FIXED_STR        6


/* Other defines    */

#define CDATA_XSTART     "![CDATA["
#define CDATA_START      "<![CDATA["
#define CDATA_END        "]]>"
#define LEN_CDATA_END    3

#define S_COMMENT        "--"      /* This is wrong, but lots of stuff incorrectly use comments: should have trailing space */
#define S_LEN_COMMENT    2
#define E_COMMENT        "-->"      /* This is wrong, but lots of stuff incorrectly use comments: should have leading space */
#define E_LEN_COMMENT    3

#define S_COND_SECT      "["
#define S_LEN_COND_SECT  1
#define E_COND_SECT      CDATA_END
#define E_LEN_COND_SECT  3
#define COND_INCLUDE     "INCLUDE"
#define COND_IGNORE      "IGNORE"

/* Doc defines */
#define INSERT_BEFORE       1
#define INSERT_AFTER        2

#define ADDCDATA_CHILD      0
#define ADDCDATA_SIBLING    1


/* System state flags and macros */

#define SYS_STATE_DTDLOADED   0x00000001
#define SYS_STATE_DTDPARSED   0x00000002
#define SYS_STATE_XMLLOADED   0x00000010
#define SYS_STATE_XMLPARSED   0x00000020
#define SYS_STATE_DOCINIT     0x00000100

#define SYS_XMLLOADED         ( SystemState & SYS_STATE_XMLLOADED )
#define SYS_DTDLOADED         ( SystemState & SYS_STATE_DTDLOADED )

#define SYS_XMLREADY(x)       ( (x & SYS_STATE_XMLLOADED) && (x & SYS_STATE_XMLPARSED) )
#define SYS_DTDREADY(x)       ( (x & SYS_STATE_DTDLOADED) && (x & SYS_STATE_DTDPARSED) )
#define SYS_SYSREADY(x)       ( (SYS_XMLREADY(x)) && (SYS_DTDREADY(x)) )
#define SYS_DOCREADY(x)       ( x & SYS_STATE_DOCINIT )

#define SYS_DOCINIT           ( SYS_DOCREADY(SystemState) )


/* Doc Make malloc()/realloc() defines */

#define MALLOC_INITIAL_SIZE    512000
#define REALLOC_CHUNK_SIZE     32768
#define ESC_CHAR_LEN           2


/* Node Copy defines */

#define COPY_NODE_ROOT         0
#define COPY_NODE_CHILD        1
#define COPY_NODE_SIBLING      2


/* Internal escape sequences */

#define NUMESCAPES       6


/* The XML class definition */

class XML
{

  private:

  struct _Node
  {
    int Type ;
    int Level ;
    long Index ;
    char *Name ;
    char *AttrList ;
    char *Value ;
    char *AttrValues ;
    struct _Node *Parent ; 
    struct _Node *FirstChild ;
    struct _Node *PreviousSibbling ;
    struct _Node *NextSibbling ;
    struct _Node *PrevIndex ;
    struct _Node *NextIndex ;
    struct _Node *Attributes ;
    int Deleted ;
  } ;

  struct _BaseDTD
  {
    int Type ;
    int Conditional ;
    int ContentType ;
    int ContentModel ;
    int Occurence ;
    int AttrType ;
    int AttrTok ;
    int AttrReq ;
    int FreeAtt ;
    long Index ;
    char *Name ;
    char *NodeData ;
    char *Attributes ;
    struct _BaseDTD *PrevIndex ;
    struct _BaseDTD *NextIndex ;
    struct _BaseDTD *FirstChild ;
    struct _BaseDTD *ContType ;
    struct _BaseDTD *ContModel ;
    struct _BaseDTD *Parent ;
    struct _BaseDTD *NextNode ;
    struct _BaseDTD *PrevNode ;
    struct _BaseDTD *FirstAttr ;
  } ;

  struct _EscapeSeqs
  {
    const char *eSeq ;
    char eChar ;
    int eLen ;
  } ;


  protected:

    int NodeLevel ;
    long Index ;
    long DTDIndex ;
    long DocIndex ;
    char *XMLData ;
    char *DTDData ;
    char *XMLVersion ;
    char *DTDVersion ;
    char *DocType ;
    char *Doc_XMLVersion ;
    char *Doc_DocType ;
    char *XMLDoc ;
    long XMLDataLength ;
    long DTDDataLength ;
    int CondLevel ;
    int DocGotDTD ;
    unsigned long SystemState ;
    struct _Node Top ;
    struct _Node *CurrentNode ;
    struct _Node Doc ;
    struct _Node *LastDocNode ;
    struct _Node *LastDocIndex ;
    struct _Node *CurrentDocNode ;
    struct _BaseDTD DTD ;
    struct _BaseDTD *CurrentDTDNode ;
    struct _Node *LastNode ;
    struct _BaseDTD *LastDTDNode ;
    struct _BaseDTD *LastDTDIndex ;
    struct _EscapeSeqs esc[NUMESCAPES] ;

    void initialise( void ) ;
    int Load( long, char *, char ** ) ;
    int ParseXML( XML * ) ;
    char *SkipWhiteSpace( char * ) ;
    char *FindWhiteSpace( char * ) ;
    char *FindClosing( char * ) ;
    char *FindEndTag( char * ) ;
    int AddNode( char *, int, XML * ) ;
    long AddDTDNode( char *, int ) ;
    long EvalCondSect( char * ) ;
    int ParseDTD1( char *, long ) ;
    int AttributeNodes( char * ) ;
    void Unescape( void ) ;
    void InitEscapes( void ) ;
    void ReplaceEscapes( void ) ;
    int ResolveDTD( void ) ;
    char *ResolveEntity( const char * ) ;
    int ResolveElement( struct _BaseDTD * ) ;
    int ResolveElementMixed( struct _BaseDTD * ) ;
    int ResolveElementKids( struct _BaseDTD * ) ;
    int ResolveElementCommon( struct _BaseDTD *, char **, char ** ) ;
    int ResolveElementDetail( struct _BaseDTD *, char *, long ) ;
    int DTDNodeContentType( char ) ;
    int ResolveAttlist( struct _BaseDTD * ) ;
    int ResolveAttlistEntities( struct _BaseDTD * ) ;
    int DTDNodeAttributeType( char *, int * ) ;
    int DTDNodeAttributeReq( char * ) ;
    int GetElementContentType( const char * ) ;
    void ZapNode( struct _BaseDTD * ) ;
    void ZapDocNode( struct _Node * ) ;
    void ZapXMLNode( struct _Node * ) ;
    char * Replace( char *, char *, char * ) ;
    struct XML::_BaseDTD *FindElementPtr( char * ) ;
    int ValidateXMLElements( XML * ) ;
    int ValidateDocElements( XML * ) ;
    int ValidateNodeAttributes( struct _BaseDTD *, XML * ) ;
    int DocValidateNodeAttributes( struct _BaseDTD *, XML * ) ;
    int DTDCheckXMLNode( XML * ) ;
    int DTDCheckDocNode( XML * ) ;
    unsigned int DocMakeLine( char *, int ) ;
    unsigned int DocMakeEndTag( char *, int ) ;
    char *EscapeSpecialChars( const char * ) ;
    int CheckRequiredElements( struct _BaseDTD * ) ;
    int ValidateParentSibling( struct _BaseDTD *, struct _BaseDTD *, struct _BaseDTD *, struct _BaseDTD *) ;
    int DocAddAttrNode( const char *, const char *, struct _Node * ) ;
    int CopyNode( int ) ;
    int CopyAttrNodes( void ) ;
    int CopyNodeCommon( void ) ;
    void DTDWipe( void ) ;
    char *CheckSpecialNodes( char *, char *, int ) ;
    char *SkipComments( char *, int * ) ;
    int ElementHasNamespace( const char * ) ;
    int NodeHasNoValue( const char * ) ;
	int NodeHasAttribute(const char *) ;
	int DocNodeHasAttribute(const char *) ;

  public:
    XML( void ) ;
    XML( long, char *, int * ) ;
    ~XML( void ) ;
    int LoadData( long, char * ) ;
    int LoadDTD( long, char * ) ;
    int Parse( void ) ;
    int ParseWithDTD( XML * ) ;
    int ParseDTD( void ) ;
    int FindNode( const char *, const long ) ;
    int FindNode( const long ) ;
    int FindNode( const int, const long ) ;
	int FindNode( const char *, const char *, const char *) ;
    int FindFirstNode( const char * ) ;
    int MoveToFirstChild( void ) ;
    int MoveToNextSibling( void ) ;
    int MoveToPrevSibling( void ) ;
    int MoveToParent( void ) ;
    int NodeHasChild( void ) ;
    int NodeHasSibling( void ) ;
    int NodeHasParent( void ) ;
    const char *NodeValue( void ) ;
    const char *NodeAttr( void ) ;
    const int NodeType( void ) ;
    const long NodeIndex( void ) ;
    void MoveToTop( void ) ;
    const char *GetVersion( void ) ;
    const char *GetDocType( void ) ;
    const char *GetNodeAttrValue( const char * ) ;
    const char *GetNodeName( void ) ;
    const char *GetParentName( void ) ;
    long GetNodeIndex( void ) ;
    long GetParentIndex( void ) ;
    unsigned long GetSystemState( void ) ;
    int ValidateXMLvsDTD( XML * ) ;
    int ValidateDocvsDTD( XML * ) ;
    int FindElement( const char * ) ;
    long GetObjectAddress( void ) ;
    int DocInitialise( const char *, XML * ) ;
    char *GetDTDElementName( void ) ;
    int GetContentType( void ) ;
    int ElementHasAttr( void ) ;
    int DocAddSibling( const char *, XML * ) ;
    int DocAddFirstChild( const char *, XML * ) ;
    int DOCAddCDATA( const char *, int ) ;
    int DocAddFrag( const char * ) ;
    int DocSetAttribute( const char *, const char *, XML *) ;
    int DocSetNodeValue( const char * ) ;
    int DocNodeHasChild( void ) ;
    int DocNodeHasSibling( void ) ;
    int DocNodeHasParent( void ) ;
    int DocNodeHasAttr( void ) ;
    const char *DocNodeValue( void ) ;
    const char *DocGetNodeName( void ) ;
    const char *DocGetParentName( void ) ;
    const int DocNodeType( void ) ;
    long DocGetNodeIndex( void ) ;
    long DocGetParentIndex( void ) ;
    const char *DocGetNodeAttrValue( const char * ) ;
    int DocMoveToFirstChild( void ) ;
    int DocMoveToNextSibling( void ) ;
    int DocMoveToPrevSibling( void ) ;
    int DocMoveToParent( void ) ;
    void DocMoveToTop( void ) ;
    int DocFindNode( const char *, const long ) ;
    int DocFindNode( const long ) ;
    int DocFindNode( const int, const long ) ;
	int DocFindNode( const char *, const char *, const char *) ;
    int DocDeleteNode( void ) ;
    void DocUnDeleteNode( void ) ;
    const int NodeDeleted( void ) ;
    void DocSetDocType( const char * ) ;
    const char *DocGetDocType( void ) ;
    int DocMake( void ) ;
    const char *GetXMLDoc( void ) ;
    void DocWipe( void ) ;
    void XMLWipe( void ) ;
    int CopyXMLToDoc( void ) ;

} ;

#endif
